import pandas as pd
import numpy as np
from dfbridge import DataframeBridge

from dash_connectivity_viewer.common.lookup_utilities import (
    get_nucleus_id_from_root_id,
    get_root_id_from_nuc_id,
    make_client
)

from .dataframe_utilities import *
from .link_utilities import voxel_resolution_from_info
from multiprocessing import cpu_count
from ..common.schema_utils import split_pt_position

def _soma_property_entry(soma_table, c):
    return {
        soma_table: {
            "root_id": c.soma_pt_root_id,
            "include": split_pt_position(c.soma_pt_position) + [c.nucleus_id_column],
            "aggregate": {
                c.num_soma_prefix: {
                    "group_by": c.soma_pt_root_id,
                    "column": c.nucleus_id_column,
                    "agg": "count",
                }
            },
            "suffix": c.num_soma_suffix,
            "table_filter": c.soma_table_query,
            "data": None,
            "data_resolution": None,
            "fill_missing": True,
        }
    }


def _synapse_properties(synapse_table, c):
    syn_props = {
        synapse_table: {
            "pre_root_id": c.pre_pt_root_id,
            "post_root_id": c.post_pt_root_id,
            "position_column": c.syn_pt_position,
            "aggregate": {
                c.num_syn_col: {
                    "column": c.syn_id_col,
                    "agg": "count",
                },
            },
        }
    }
    syn_props[synapse_table]["aggregate"].update(c.synapse_aggregation_rules)
    return syn_props


class NeuronData(object):
    def __init__(
        self,
        object_id,
        client,
        config,
        property_tables={},
        timestamp=None,
        n_threads=None,
        id_type="root",
        is_live=True,
    ):

        if id_type == "root":
            self._root_id = object_id
            self._nucleus_id = None
        elif id_type == "nucleus":
            self._root_id = None
            self._nucleus_id = object_id

        self._client = make_client(
            datastack=client.datastack_name,
            server_address=client.server_address,
            pool_block=True,
            pool_maxsize=config.pool_maxsize,
        )

        self._property_tables = property_tables

        if config.synapse_table is None:
            synapse_table = client.info.get_datastack_info().get("synapse_table")
        self._synapse_table = synapse_table
        self._synapse_table_properties = _synapse_properties(synapse_table, config)

        if config.soma_table is None:
            soma_table = client.info.get_datastack_info().get("soma_table")
        else:
            soma_table = config.soma_table

        self._soma_table = soma_table
        self.config = config

        self._timestamp = timestamp

        self.is_live = is_live

        self._pre_syn_df = None
        self._post_syn_df = None
        self._synapse_data_resolution = np.array([1,1,1])

        self._viewer_resolution = voxel_resolution_from_info(client.info.info_cache)

        if n_threads is None:
            n_threads = cpu_count()
        self.n_threads = n_threads

        self._partner_soma_table = None
        self._partner_root_ids = None

        self.check_root_id()
        if soma_table is not None:
            self._property_tables.update(
                _soma_property_entry(
                    soma_table,
                    config,
                )
            )

    @property
    def root_id(self):
        if self._root_id is None:
            new_root_id = get_root_id_from_nuc_id(
                self._nucleus_id,
                self.client,
                self.soma_table,
                self.config,
                self.timestamp,
                self.is_live,
            )
            if new_root_id is None:
                raise Exception("Nucleus ID not found in soma table")
            else:
                self._root_id = new_root_id
        return self._root_id

    def check_root_id(self):
        if self.client.chunkedgraph.is_latest_roots([self.root_id])[0]:
            pass
        else:
            raise Exception(f"Root id is not valid at this timestamp.")

    @property
    def nucleus_id(self):
        if self.soma_table is None:
            return None
        if self._nucleus_id is None:
            self._nucleus_id = get_nucleus_id_from_root_id(
                self._root_id,
                self.client,
                self.soma_table,
                self.config,
                self.timestamp,
            )
        return self._nucleus_id

    @property
    def client(self):
        return self._client

    @property
    def live_query(self):
        return self.is_live

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def synapse_table(self):
        return self._synapse_table

    @property
    def soma_table(self):
        return self._soma_table

    @property
    def synapse_data_resolution(self):
        return self._synapse_data_resolution

    @property
    def property_tables(self):
        return [k for k in self._property_tables]

    def pre_syn_df(self):
        if self._pre_syn_df is None:
            self._get_syn_df()
        return self._pre_syn_df.copy()

    def post_syn_df(self):
        if self._post_syn_df is None:
            self._get_syn_df()
        return self._post_syn_df.copy()

    def _get_syn_df(self):
        self._pre_syn_df, self._post_syn_df = synapse_data(
            synapse_table=self.synapse_table,
            root_id=self.root_id,
            client=self.client,
            timestamp=self.timestamp,
            config=self.config,
            n_threads=self.n_threads,
            is_live=self.is_live,
        )
        self._populate_property_tables()

    @property
    def partner_root_ids(self):
        if self._partner_root_ids is None:
            self._partner_root_ids = self._populate_root_ids()
        return self._partner_root_ids

    def _populate_root_ids(self):
        if self._pre_syn_df is None:
            self._get_syn_df()

        root_ids = np.unique(
            np.concatenate(
                (
                    self._pre_syn_df[self.config.post_pt_root_id].values,
                    self._post_syn_df[self.config.pre_pt_root_id].values,
                )
            )
        )
        return root_ids[root_ids!=0]

    def partners_out(self, properties=True):
        return self._targ_table("pre", properties)

    def partners_in(self, properties=True):
        return self._targ_table("post", properties)

    def _targ_table(self, side, properties):
        if side == "pre":
            prefix = "post"
            syn_df_grp = self.pre_syn_df().groupby(f"{prefix}_pt_root_id")
        elif side == "post":
            prefix = "pre"
            syn_df_grp = self.post_syn_df().groupby(f"{prefix}_pt_root_id")
        targ_df = self._make_simple_targ_df(syn_df_grp).rename(
            columns={f"{prefix}_pt_root_id": self.config.root_id_col}
        )
        if properties:
            targ_df = self._merge_property_tables(targ_df, self.config.root_id_col)
        for cn in self.config.target_table_display:
            if cn not in targ_df.columns:
                targ_df[cn] = np.nan
        return targ_df

    def _make_simple_targ_df(self, df_grp):
        num_syn = df_grp[self.config.syn_pt_position_split[0]].agg(len)
        syn_data_dict = {}
        for k in self.config.syn_pt_position_split:
            syn_data_dict[k] = df_grp[k].agg(list)
        syn_data_dict[self.config.num_syn_col] = num_syn
        syn_df = pd.DataFrame(syn_data_dict)

        for k, v in self.config.synapse_aggregation_rules.items():
            syn_df[k] = df_grp[v["column"]].agg(v["agg"])
        return syn_df.sort_values(
            by=self.config.num_syn_col, ascending=False
        ).reset_index()

    def _populate_property_tables(self):
        dfs = property_table_data(
            self.partner_root_ids,
            self._property_tables,
            self.client,
            self.timestamp,
            self.n_threads,
            self.is_live,
        )
        for k, df in dfs.items():
            dbf = DataframeBridge(
                self._property_tables[k].get("table_bridge_schema", None)
            )
            self._property_tables[k]["data"] = dbf.reformat(df).fillna(np.nan)
            self._property_tables[k]["data_resolution"] = DESIRED_RESOLUTION

    def property_data(self, table_name):
        if self._property_tables.get(table_name).get("data") is None:
            self._populate_property_tables()
        return self._property_tables.get(table_name, {}).get("data")

    def property_data_resolution(self, table_name):
        if self._property_tables.get(table_name).get("data") is None:
            self._populate_property_tables()
        return self._property_tables.get(table_name).get("data_resolution")

    def property_root_id_column(self, table_name):
        return self._property_tables.get(table_name).get("root_id")

    def _agg_columns(self, table_name):
        return list(
            self._property_tables.get(table_name, {}).get("aggregate", {}).keys()
        )

    def property_columns(self, table_name):
        return (
            [self.property_root_id_column(table_name)]
            + self._property_tables.get(table_name).get("include")
            + self._agg_columns(table_name)
        )

    def property_column_suffix(self, table_name):
        return self._property_tables.get(table_name).get("suffix", "")

    def _merge_property_tables(self, df, merge_column):
        for tn in self.property_tables:
            df = df.merge(
                self.property_data(tn),
                left_on=merge_column,
                right_index=True,
                how="left",
                suffixes=("", self.property_column_suffix(tn)),
            )
            df.rename(
                columns={
                    c: f"{c}{self.property_column_suffix(tn)}"
                    for c in self.property_columns(tn)
                    if f"{c}{self.property_column_suffix(tn)}" not in df.columns
                },
                inplace=True,
            )

        if self.soma_table is not None:
            df[self.config.num_soma_col] = (
                df[self.config.num_soma_col].fillna(0).astype(int)
            )

        return df

    def _get_own_soma_loc(self):
        own_soma_df = get_specific_soma(
            self.soma_table,
            self.root_id,
            self.client,
            self.timestamp,
            self.is_live,
        )
        if len(own_soma_df) != 1:
            own_soma_loc = np.nan
        else:
            own_soma_loc = own_soma_df[self.config.soma_pt_position].values[0]
        return own_soma_loc

    def syn_all_df(self):
        pre_df = self.pre_syn_df()
        pre_df["direction"] = "pre"
        post_df = self.post_syn_df()
        post_df["direction"] = "post"

        syn_df = pd.concat([pre_df, post_df])
        syn_df["x"] = 0

        return syn_df

    def soma_location(self):
        if self.soma_table is None:
            return None
        return np.array(self._get_own_soma_loc())

    def soma_location_list(self, length):
        return np.repeat(np.atleast_2d(self.soma_location()), length, axis=0).tolist()



from .config import *


def process_dataframe(df, root_id_column, pt_column):
    df["num_anno"] = df.groupby(root_id_column).transform("count")[
        f"{bound_pt_position(pt_column)}_x"
    ]
    return df



import plotly.graph_objects as go
from plotly.subplots import make_subplots
from .cortex_plots import *

bg_color = "White"
plotly_template = "plotly_white"

def bar_fig_df(df, config, color_column, width=450, height=350):
    bar = bar_plot_df(df, config, color_column)

    fig = go.Figure()
    fig.add_trace(bar)
        
    fig.update_layout(
        autosize=True,
        height=height,
        width=width,
        paper_bgcolor=bg_color,
        template=plotly_template,
        showlegend=False,
        margin=dict(l=20, r=20, t=20, b=20),
    )

    return fig


def violin_fig(ndat, height=350, width=200):

    fig = go.Figure()

    violin_post = post_violin_plot(ndat)
    violin_pre = pre_violin_plot(ndat)
    fig.add_trace(violin_post)
    fig.add_trace(violin_pre)

    fig.update_layout(
        yaxis_title="Synapse Depth",
        height=height,
        width=width,
        paper_bgcolor=bg_color,
        template=plotly_template,
        showlegend=False,
        margin=dict(l=40, r=20, t=20, b=20),
    )

    fig.update_yaxes(
        tickvals=ndat.config.vis.ticklocs,
        ticktext=ndat.config.vis.tick_labels,
        ticklabelposition="outside bottom",
        range=ndat.config.height_bnds.astype(int)[::-1].tolist(),
        gridcolor="#CCC",
        gridwidth=2,
    )
    return fig

def scatter_fig_df(df, config, color_column, width=350, height=350):
    fig = go.Figure()
    scatter = synapse_soma_scatterplot(df, config, color_column)
    fig.add_traces(scatter)

    fig.update_layout(
        xaxis_title="Soma Depth",
        yaxis_title="Synapse Depth",
        height=height,
        width=width,
        paper_bgcolor=bg_color,
        template=plotly_template,
        showlegend=True,
        margin=dict(l=20, r=20, t=20, b=20),
    )

    fig.update_xaxes(
        tickvals=config.vis.ticklocs,
        ticktext=config.vis.tick_labels,
        ticklabelposition="outside right",
        gridcolor="#CCC",
        gridwidth=2,
        scaleanchor="y",
    )

    fig.update_yaxes(
        tickvals=config.vis.ticklocs,
        ticktext=config.vis.tick_labels,
        ticklabelposition="outside bottom",
        range=config.height_bnds.astype(int)[::-1].tolist(),
        gridcolor="#CCC",
        gridwidth=2,
        scaleanchor="x",
        scaleratio=1,
    )
    return fig


def scatter_fig(ndat, color_column, width=350, height=350):

    fig = go.Figure()
    scatter = synapse_soma_scatterplot(
        ndat,
        ndat.config.synapse_depth_column,
        ndat.config.soma_depth_column,
        color_column,
    )
    fig.add_traces(scatter)

    fig.update_layout(
        xaxis_title="Soma Depth",
        yaxis_title="Synapse Depth",
        height=height,
        width=width,
        paper_bgcolor="White",
        template="plotly_white",
        showlegend=True,
        margin=dict(l=20, r=20, t=20, b=20),
    )

    fig.update_xaxes(
        tickvals=ndat.config.vis.ticklocs,
        ticktext=ndat.config.vis.tick_labels,
        ticklabelposition="outside right",
        gridcolor="#CCC",
        gridwidth=2,
        scaleanchor="y",
    )

    fig.update_yaxes(
        tickvals=ndat.config.vis.ticklocs,
        ticktext=ndat.config.vis.tick_labels,
        ticklabelposition="outside bottom",
        range=ndat.config.height_bnds.astype(int)[::-1].tolist(),
        gridcolor="#CCC",
        gridwidth=2,
        scaleanchor="x",
        scaleratio=1,
    )
    return fig



from logging import info
from nglui import statebuilder
import pandas as pd
import numpy as np
from seaborn import color_palette
from itertools import cycle
from .lookup_utilities import make_client
from .schema_utils import bound_pt_position, bound_pt_root_id
from .dataframe_utilities import rehydrate_dataframe

EMPTY_INFO_CACHE = {"aligned_volume": {}, "cell_type_column": None}
MAX_URL_LENGTH = 1_750_000
DEFAULT_NGL = "https://neuromancer-seung-import.appspot.com/"


def image_source(info_cache):
    if info_cache is None:
        return None
    return info_cache["aligned_volume"].get("image_source", "")

def aligned_volume(info_cache):
    if info_cache is None:
        return None
    return info_cache.get('aligned_volume', {}).get('name')

def seg_source(info_cache):
    if info_cache is None:
        return None
    return info_cache.get("segmentation_source", "")


def viewer_site(info_cache):
    if info_cache is None:
        return None
    return info_cache.get("viewer_site", "")


def state_server(info_cache):
    if info_cache is None:
        return None
    return f"{info_cache.get('global_server', '')}/nglstate/api/v1/post"


def root_id(info_cache):
    if info_cache is None:
        return None
    return int(info_cache.get("root_id", None))


def timestamp(info_cache):
    if info_cache is None:
        return None
    return info_cache.get("ngl_timestamp", None)


def voxel_resolution_from_info(info_cache):
    try:
        vr = [
            info_cache.get("viewer_resolution_x"),
            info_cache.get("viewer_resolution_y"),
            info_cache.get("viewer_resolution_z"),
        ]
        return vr
    except:
        return None


def statebuilder_kwargs(info_cache):
    return dict(
        url_prefix=viewer_site(info_cache),
        state_server=state_server(info_cache),
        resolution=voxel_resolution_from_info(info_cache),
    )


def generate_statebuilder(
    info_cache,
    config,
    base_root_id=None,
    base_color="#ffffff",
    preselect_all=True,
    anno_column="post_pt_root_id",
    anno_layer="syns",
    data_resolution=[1,1,1],
):
    img = statebuilder.ImageLayerConfig(
        image_source(info_cache),
        contrast_controls=True,
        black=config.image_black,
        white=config.image_white,
    )
    if preselect_all:
        selected_ids_column = [anno_column]
    else:
        selected_ids_column = None
    if base_root_id is None:
        base_root_id = []
        base_color = [None]
    else:
        base_root_id = [base_root_id]
        base_color = [base_color]

    seg = statebuilder.SegmentationLayerConfig(
        seg_source(info_cache),
        selected_ids_column=selected_ids_column,
        fixed_ids=base_root_id,
        fixed_id_colors=base_color,
        alpha_3d=0.8,
        timestamp=timestamp(info_cache),
    )

    points = statebuilder.PointMapper(
        config.syn_pt_position,
        linked_segmentation_column=anno_column,
        group_column=anno_column,
        multipoint=True,
        set_position=True,
        collapse_groups=True,
        split_positions=True,
    )
    anno = statebuilder.AnnotationLayerConfig(
        anno_layer,
        mapping_rules=points,
        linked_segmentation_layer=seg.name,
        filter_by_segmentation=True,
        data_resolution=data_resolution,
    )

    sb = statebuilder.StateBuilder(
        [img, seg, anno],
        **statebuilder_kwargs(info_cache),
    )
    return sb


def generate_statebuilder_pre(
    info_cache, config, preselect=False, data_resolution=[1,1,1],
):

    img = statebuilder.ImageLayerConfig(
        image_source(info_cache),
        contrast_controls=True,
        black=config.image_black,
        white=config.image_white,
    )
    seg = statebuilder.SegmentationLayerConfig(
        seg_source(info_cache),
        fixed_ids=[root_id(info_cache)],
        fixed_id_colors=["#ffffff"],
        alpha_3d=0.8,
        timestamp=timestamp(info_cache),
    )
    points = statebuilder.PointMapper(
        config.syn_pt_position,
        linked_segmentation_column=config.root_id_col,
        set_position=True,
        multipoint=True,
        split_positions=True,
    )
    anno = statebuilder.AnnotationLayerConfig(
        "output_syns",
        mapping_rules=points,
        linked_segmentation_layer=seg.name,
        data_resolution=data_resolution,
    )
    sb = statebuilder.StateBuilder(
        [img, seg, anno],
        **statebuilder_kwargs(info_cache),
    )
    return sb


def generate_statebuilder_post(info_cache, config, data_resolution=[1,1,1]):
    img = statebuilder.ImageLayerConfig(
        image_source(info_cache),
        contrast_controls=True,
        black=config.image_black,
        white=config.image_white,
    )

    seg = statebuilder.SegmentationLayerConfig(
        seg_source(info_cache),
        fixed_ids=[root_id(info_cache)],
        fixed_id_colors=["#ffffff"],
        alpha_3d=0.8,
        timestamp=timestamp(info_cache),
    )
    points = statebuilder.PointMapper(
        config.syn_pt_position,
        linked_segmentation_column=config.root_id_col,
        set_position=True,
        split_positions=True,
        multipoint=True,
    )
    anno = statebuilder.AnnotationLayerConfig(
        "input_syns",
        mapping_rules=points,
        linked_segmentation_layer=seg.name,
        data_resolution=data_resolution,
    )
    sb = statebuilder.StateBuilder(
        [img, seg, anno],
        **statebuilder_kwargs(info_cache),
    )
    return sb


def generate_statebuider_syn_grouped(
    info_cache,
    anno_name,
    config,
    fixed_id_color="#FFFFFF",
    preselect=False,
    data_resolution=[1,1,1],
):
    points = statebuilder.PointMapper(
        point_column=config.syn_pt_position,
        linked_segmentation_column=config.root_id_col,
        group_column=config.root_id_col,
        split_positions=True,
        multipoint=True,
        set_position=True,
        collapse_groups=True,
    )

    img = statebuilder.ImageLayerConfig(
        image_source(info_cache),
        contrast_controls=True,
        black=config.image_black,
        white=config.image_white,
    )

    if preselect:
        selected_ids_column = config.root_id_col
    else:
        selected_ids_column = None

    seg = statebuilder.SegmentationLayerConfig(
        seg_source(info_cache),
        fixed_ids=[root_id(info_cache)],
        fixed_id_colors=[fixed_id_color],
        selected_ids_column=selected_ids_column,
        alpha_3d=0.8,
        timestamp=timestamp(info_cache),
    )

    anno = statebuilder.AnnotationLayerConfig(
        anno_name,
        mapping_rules=points,
        linked_segmentation_layer=seg.name,
        filter_by_segmentation=True,
        data_resolution=data_resolution,
    )

    sb = statebuilder.StateBuilder(
        [img, seg, anno],
        **statebuilder_kwargs(info_cache),
    )

    return sb


def generate_url_cell_types(
    selected_rows,
    df,
    info_cache,
    config,
    pt_column,
    cell_type_column='cell_type',
    group_annotations=False,
    multipoint=False,
    fill_null=None,
    return_as="url",
    data_resolution=[1,1,1],
):
    
    if len(selected_rows) > 0 or selected_rows is None:
        df = df.iloc[selected_rows].reset_index(drop=True)

    img = statebuilder.ImageLayerConfig(
        image_source(info_cache),
        contrast_controls=True,
        black=config.image_black,
        white=config.image_white,
    )
    seg = statebuilder.SegmentationLayerConfig(
        seg_source(info_cache),
        alpha_3d=0.8,
        timestamp=timestamp(info_cache),
    )
    # sbs = [
    #     statebuilder.StateBuilder(
    #         [img, seg],
    #         **statebuilder_kwargs(info_cache),
    #     )
    # ]
    annos = []
    if group_annotations:
        if fill_null:
            df[cell_type_column].cat.add_categories(fill_null, inplace=True)
            df[cell_type_column].fillna(fill_null, inplace=True)
        cell_types = np.sort(pd.unique(df[cell_type_column].dropna()))
        colors = color_palette("tab20").as_hex()

        for ct, clr in zip(cell_types, cycle(colors)):
            annos.append(
                statebuilder.AnnotationLayerConfig(
                    ct,
                    color=clr,
                    linked_segmentation_layer=seg.name,
                    data_resolution=data_resolution,
                    mapping_rules=statebuilder.PointMapper(
                        bound_pt_position(pt_column),
                        linked_segmentation_column=config.root_id_col,
                        set_position=True,
                        multipoint=multipoint,
                        split_positions=True,
                        mapping_set=ct,
                    )
                )
            )
        sb = statebuilder.StateBuilder([img, seg] + annos, **statebuilder_kwargs(info_cache))
        return sb.render_state(
            {ct: df.query(f"{cell_type_column}==@ct") for ct in cell_types},
            return_as=return_as,
        )
    else:
        if cell_type_column is not None:
            if len(cell_type_column)==0:
                cell_type_column=None
        anno = statebuilder.AnnotationLayerConfig(
            'Annotations',
            linked_segmentation_layer=seg.name,
            mapping_rules=statebuilder.PointMapper(
                bound_pt_position(pt_column),
                linked_segmentation_column=config.root_id_col,
                split_positions=True,
                multipoint=multipoint,
                set_position=True,
                description_column=cell_type_column,
            ),
            data_resolution=data_resolution,
        )
        sb = statebuilder.StateBuilder([img, seg, anno], **statebuilder_kwargs(info_cache))
        return sb.render_state(
            df,
            return_as=return_as,
        )

    # for ct, clr in zip(cell_types, cycle(colors)):
    #     anno = statebuilder.AnnotationLayerConfig(
    #         ct,
    #         color=clr,
    #         linked_segmentation_layer=seg.name,
    #         mapping_rules=statebuilder.PointMapper(
    #             bound_pt_position(pt_column),
    #             linked_segmentation_column=config.root_id_col,
    #             set_position=True,
    #             multipoint=multipoint,
    #             split_positions=True,
    #         ),
    #         data_resolution=data_resolution,
    #     )
    #     sbs.append(
    #         statebuilder.StateBuilder(
    #             [anno],
    #             **statebuilder_kwargs(info_cache),
    #         )
    #     )
    #     dfs.append(df.query("cell_type == @ct"))
    # csb = statebuilder.ChainedStateBuilder(sbs)
    # return csb.render_state(dfs, return_as=return_as)


def generate_statebuilder_syn_cell_types(
    info_cache,
    rows,
    config,
    cell_type_column="cell_type",
    group_annotations=True,
    multipoint=False,
    fill_null=None,
    data_resolution=[1,1,1],
    include_no_type=True,
):
    df = rehydrate_dataframe(rows, config.syn_pt_position_split)
    if fill_null and include_no_type:
        df[cell_type_column].fillna(fill_null, inplace=True)

    cell_types = np.sort(pd.unique(df[cell_type_column].dropna()))
    img = statebuilder.ImageLayerConfig(
        image_source(info_cache),
        contrast_controls=True,
        black=config.image_black,
        white=config.image_white,
    )
    seg = statebuilder.SegmentationLayerConfig(
        seg_source(info_cache),
        alpha_3d=0.8,
        fixed_ids=[int(info_cache["root_id"])],
        timestamp=timestamp(info_cache),
    )

    # sbs = [
    #     statebuilder.StateBuilder(
    #         [img, seg],
    #         **statebuilder_kwargs(info_cache),
    #     )
    # ]
    colors = color_palette("tab20").as_hex()
    annos = []
    for ct, clr in zip(cell_types, cycle(colors)):
        anno = statebuilder.AnnotationLayerConfig(
            ct,
            color=clr,
            linked_segmentation_layer=seg.name,
            mapping_rules=statebuilder.PointMapper(
                config.syn_pt_position,
                linked_segmentation_column=config.root_id_col,
                set_position=True,
                multipoint=multipoint,
                split_positions=True,
                mapping_set=ct,
            ),
            data_resolution=data_resolution,
        )
        annos.append(anno)
            # statebuilder.StateBuilder(
                # [anno],
                # **statebuilder_kwargs(info_cache),
            # )
        # )
        # dfs.append(df.query(f"{cell_type_column} == @ct"))
    sb = statebuilder.StateBuilder([img, seg] + annos, **statebuilder_kwargs(info_cache))
    # csb = statebuilder.ChainedStateBuilder(sbs)
    df_dict = {ct: df.query(f'{cell_type_column}=="{ct}"') for ct in cell_types}
    return sb, df_dict


def make_url_robust(df, sb, datastack, config):
    """Generate a url from a neuroglancer state. If too long, return through state server"""
    url = sb.render_state(df, return_as="url")
    if len(url) > MAX_URL_LENGTH:
        client = make_client(datastack, config.server_address)
        state = sb.render_state(df, return_as="dict")
        state_id = client.state.upload_state_json(state)
        ngl_url = client.info.viewer_site()
        if ngl_url is None:
            ngl_url = DEFAULT_NGL
        url = client.state.build_neuroglancer_url(state_id, ngl_url=ngl_url)
    return url



import pandas as pd
import numpy as np
from .lookup_utilities import get_root_id_from_nuc_id, make_client
from .dataframe_utilities import query_table_any
from .schema_utils import get_table_info
from ..common.config import RegisterTable
from dfbridge import DataframeBridge
from copy import copy


def _table_schema(config):
    base_schema = {config.root_id_col: config.ct_cell_type_root_id}
    for ptx, sptx in zip(
        ["pt_position_x", "pt_position_y", "pt_position_z"],
        config.ct_cell_type_pt_position_split,
    ):
        base_schema[ptx] = sptx
    for k in config.value_columns:
        base_schema[k] = k
    return base_schema


class TableViewer(object):
    def __init__(
        self,
        table_name,
        client,
        config,
        timestamp=None,
        id_query=None,
        id_query_type=None,
        column_query={},
        is_live=True,
    ):

        self._client = make_client(
            datastack=client.datastack_name,
            server_address=client.server_address,
        )
        pt, add_cols = get_table_info(table_name, self._client)
        config = RegisterTable(pt, add_cols, config)
        self.config = config
        self._cell_type_bridge_schema = _table_schema(config)

        if config.soma_table is None:
            soma_table = self._client.info.get_datastack_info().get("soma_table")

        self._soma_table = soma_table
        self._table_name = table_name

        self._data = None
        self._data_resolution = config.data_resolution

        self._column_query = column_query
        self._annotation_query = None
        self._id_query = id_query
        self._id_query_type = id_query_type
        self._timestamp = timestamp
        self.is_live = is_live

        self._process_id_query()

    @property
    def client(self):
        return self._client

    @property
    def live_query(self):
        return self.is_live

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def table_name(self):
        return self._table_name

    @property
    def soma_table(self):
        return self._soma_table

    def table_data(self):
        if self._data is None:
            self._populate_data()
        return self._data

    @property
    def table_resolution(self):
        return self._data_resolution

    @property
    def cell_type_bridge(self):
        return DataframeBridge(self._cell_type_bridge_schema)

    def _populate_data(self):
        id_column = None
        ids = None
        if self._id_query is not None:
            id_column = self.config.ct_cell_type_root_id
            ids = self._id_query
        if self._annotation_query is not None:
            id_column = "id"
            ids = self._annotation_query

        df = query_table_any(
            self.table_name,
            id_column,
            ids,
            self.client,
            self.timestamp,
            self._column_query,
            is_live=self.is_live,
        )
        self._data = self.cell_type_bridge.reformat(df).fillna(np.nan)

    def _process_id_query(self):
        if self._id_query_type == "root":
            self._id_query = self._id_query
            self._annotation_query = None
        elif self._id_query_type == "nucleus":
            self._id_query = self._lookup_roots_from_nucleus(self._id_query)
            self._annotation_query = None
        elif self._id_query_type == "annotation":
            self._annotation_query = copy(self._id_query)
            self._id_query = None

    def _lookup_roots_from_nucleus(self, soma_ids):
        df = self.client.materialize.query_table(
            self.soma_table,
            filter_in_dict={self.config.nucleus_id_column: soma_ids},
            timestamp=self.timestamp,
        )
        if self.config.soma_table_query is not None:
            df = df.query(self.config.soma_table_query)
        return df[self.config.soma_pt_root_id].values



from dash import html
from ..common.neuron_data_base import NeuronData
from dash.dependencies import Input, Output, State
from dash import callback_context

from ..common.link_utilities import (
    generate_statebuider_syn_grouped,
    generate_statebuilder,
    generate_statebuilder_pre,
    generate_statebuilder_post,
    EMPTY_INFO_CACHE,
    MAX_URL_LENGTH,
    make_url_robust,
)
from ..common.dataframe_utilities import (
    stringify_root_ids, stringify_list, repopulate_list
)
from ..common.dash_url_helper import _COMPONENT_ID_TYPE
from ..common.lookup_utilities import make_client
from .config import ConnectivityConfig

import datetime
import pandas as pd

try:
    from loguru import logger
    import time
except:
    logger = None


InputDatastack = Input({"id_inner": "datastack", "type": _COMPONENT_ID_TYPE}, "value")
OutputDatastack = Output({"id_inner": "datastack", "type": _COMPONENT_ID_TYPE}, "value")
StateAnnoID = State({"id_inner": "anno-id", "type": _COMPONENT_ID_TYPE}, "value")
StateAnnoType = State({"id_inner": "cell-id-type", "type": _COMPONENT_ID_TYPE}, "value")
StateLiveQuery = State(
    {"id_inner": "live-query-toggle", "type": _COMPONENT_ID_TYPE}, "value"
)

OutputLiveQueryToggle = Output(
    {"id_inner": "live-query-toggle", "type": _COMPONENT_ID_TYPE},
    "options",
)
OutputLiveQueryValue = Output(
    {"id_inner": "live-query-toggle", "type": _COMPONENT_ID_TYPE}, "value"
)


def register_callbacks(app, config):
    c = ConnectivityConfig(config)

    @app.callback(
        Output("data-table", "selected_rows"),
        Input("reset-selection", "n_clicks"),
        Input("connectivity-tab", "value"),
    )
    def reset_selection(n_clicks, tab_value):
        return []

    @app.callback(
        Output("header-bar", 'children'),
        InputDatastack,
    )
    def set_header(datastack):
        return html.H3(f"Connectivity Info — {datastack}", className="bg-primary text-white p-2 mb-2 text-center")

    @app.callback(
        Output("data-table", "columns"),
        InputDatastack,
    )
    def define_table_columns(_):
        return [{"name": i, "id": i} for i in c.table_columns]

    @app.callback(
        OutputDatastack,
        InputDatastack,
    )
    def define_datastack(datastack):
        if datastack is None:
            datastack = ""

        if len(datastack) == 0:
            return c.default_datastack
        else:
            return datastack

    @app.callback(
        OutputLiveQueryToggle,
        OutputLiveQueryValue,
        InputDatastack,
        StateLiveQuery,
    )
    def disable_live_query(_, lq):
        options_active = [{"label": "Live Query", "value": 1}]
        options_disabled = [{"label": "Live Query", "value": 1, "disabled": True}]
        if c.disallow_live_query:
            return options_disabled, ""
        else:
            return options_active, lq

    @app.callback(
        Output("target-table-json", "data"),
        Output("source-table-json", "data"),
        Output("output-tab", "label"),
        Output("input-tab", "label"),
        Output("reset-selection", "n_clicks"),
        Output("client-info-json", "data"),
        Output("loading-spinner", "children"),
        Output("message-text", "children"),
        Output("message-text", "color"),
        Output("synapse-table-resolution-json", "data"),
        Input("submit-button", "n_clicks"),
        InputDatastack,
        StateAnnoID,
        StateAnnoType,
        StateLiveQuery,
    )
    def update_data(_, datastack_name, anno_id, id_type, live_query_toggle):
        if logger is not None:
            t0 = time.time()

        try:
            client = make_client(datastack_name, c.server_address)
            info_cache = client.info.info_cache[datastack_name]
            info_cache["global_server"] = client.server_address
        except Exception as e:
            return (
                [],
                [],
                "Output",
                "Input",
                1,
                EMPTY_INFO_CACHE,
                "",
                str(e),
                "danger",
                None,
            )

        if len(anno_id) == 0:
            return (
                [],
                [],
                "Output",
                "Input",
                1,
                info_cache,
                "",
                "No annotation id selected",
                "info",
                None,
            )

        if len(anno_id) == 0:
            anno_id = None
            id_type = "anno_id"

        live_query = len(live_query_toggle) == 1
        if live_query and not c.disallow_live_query:
            timestamp = datetime.datetime.utcnow()
        else:
            timestamp = None
            timestamp_ngl = client.materialize.get_timestamp()
            info_cache["ngl_timestamp"] = timestamp_ngl.timestamp()

        if anno_id is None:
            root_id = None
        else:
            if id_type == "root_id":
                object_id = int(anno_id)
                object_id_type = "root"
            elif id_type == "nucleus_id":
                object_id = int(anno_id)
                object_id_type = "nucleus"
            else:
                raise ValueError('id_type must be either "root_id" or "nucleus_id"')

        try:
            nrn_data = NeuronData(
                object_id,
                client,
                config=c,
                timestamp=timestamp,
                id_type=object_id_type,
                n_threads=1,
            )

            root_id = nrn_data.root_id

            pre_targ_df = nrn_data.partners_out()
            pre_targ_df = stringify_root_ids(
                pre_targ_df, stringify_cols=[c.root_id_col]
            )

            post_targ_df = nrn_data.partners_in()
            post_targ_df = stringify_root_ids(
                post_targ_df, stringify_cols=[c.root_id_col]
            )
            for col in nrn_data.config.syn_pt_position_split:
                stringify_list(col, pre_targ_df)
                stringify_list(col, post_targ_df)

            n_syn_pre = pre_targ_df[c.num_syn_col].sum()
            n_syn_post = post_targ_df[c.num_syn_col].sum()

            info_cache["root_id"] = str(root_id)

        except Exception as e:
            return (
                [],
                [],
                "Output",
                "Input",
                1,
                EMPTY_INFO_CACHE,
                "",
                str(e),
                "danger",
                None,
            )

        if logger is not None:
            logger.info(
                f"Data update for {root_id} | time:{time.time() - t0:.2f} s, syn_in: {n_syn_post} , syn_out: {n_syn_pre}"
            )

        if timestamp is not None:
            output_message = f"Current connectivity for root id {root_id}"
            output_status = "success"
        else:
            output_message = f"Connectivity for root id {root_id} materialized on {timestamp_ngl:%m/%d/%Y} (v{client.materialize.version})"
            output_status = "success"

        return (
            pre_targ_df.to_dict("records"),
            post_targ_df.to_dict("records"),
            f"Output (n = {n_syn_pre})",
            f"Input (n = {n_syn_post})",
            1,
            info_cache,
            "",
            output_message,
            output_status,
            nrn_data.synapse_data_resolution,
        )

    @app.callback(
        Output("data-table", "data"),
        Input("connectivity-tab", "value"),
        Input("target-table-json", "data"),
        Input("source-table-json", "data"),
    )
    def update_table(
        tab_value,
        pre_data,
        post_data,
    ):
        if tab_value == "tab-pre":
            return pre_data
        elif tab_value == "tab-post":
            return post_data
        else:
            return []

    @app.callback(
        Output("ngl_link", "href"),
        Output("ngl_link", "children"),
        Output("ngl_link", "disabled"),
        Output("link-loading", "children"),
        Input("connectivity-tab", "value"),
        Input("data-table", "derived_virtual_data"),
        Input("data-table", "derived_virtual_selected_rows"),
        Input("client-info-json", "data"),
        Input("synapse-table-resolution-json", "data"),
    )
    def update_link(
        tab_value,
        rows,
        selected_rows,
        info_cache,
        data_resolution,
    ):
        large_state_text = "State Too Large - Please Filter"
        def small_state_text(n):
            return f"Neuroglancer: ({n} partners)"

        if info_cache is None:
            return "", "No datastack set", True, ""

        if rows is None or len(rows) == 0:
            rows = {}
            sb = generate_statebuilder(info_cache, c)
            return (
                sb.render_state(None, return_as="url"),
                small_state_text(0),
                False,
                "",
            )
        else:
            syn_df = pd.DataFrame(rows)
            for col in c.syn_pt_position_split:
                repopulate_list(col, syn_df)
            if len(selected_rows) == 0:
                if tab_value == "tab-pre":
                    sb = generate_statebuilder_pre(
                        info_cache, c, data_resolution=data_resolution
                    )
                elif tab_value == "tab-post":
                    sb = generate_statebuilder_post(
                        info_cache, c, data_resolution=data_resolution
                    )
                else:
                    raise ValueError('tab must be "tab-pre" or "tab-post"')
                url = sb.render_state(
                    syn_df.sort_values(by=c.num_syn_col, ascending=False),
                    return_as="url",
                )
                small_out_text = small_state_text(len(syn_df))
            else:
                if tab_value == "tab-pre":
                    anno_layer = "Output Synapses"
                elif tab_value == "tab-post":
                    anno_layer = "Input Synapses"
                sb = generate_statebuider_syn_grouped(
                    info_cache,
                    anno_layer,
                    c,
                    preselect=len(selected_rows) == 1,
                    data_resolution=data_resolution,
                )
                url = sb.render_state(syn_df.iloc[selected_rows], return_as="url")
                small_out_text = small_state_text(len(selected_rows))

        if len(url) > MAX_URL_LENGTH:
            return "", large_state_text, True, ""
        else:
            return url, small_out_text, False, ""

    @app.callback(
        Output("all-input-link", "children"),
        Input("all-input-link-button", "n_clicks"),
        Input("submit-button", "n_clicks"),
        Input("source-table-json", "data"),
        Input("client-info-json", "data"),
        InputDatastack,
        Input("synapse-table-resolution-json", "data"),
        prevent_initial_call=True,
    )
    def generate_all_input_link(_1, _2, rows, info_cache, datastack, data_resolution):
        ctx = callback_context
        if not ctx.triggered:
            return ""
        trigger_src = ctx.triggered[0]["prop_id"].split(".")[0]
        if (
            trigger_src == "submit-button"
            or trigger_src == "client-info-json"
            or trigger_src == "source-table-json"
        ):
            return ""

        if rows is None or len(rows) == 0:
            return html.Div("No inputs to show")
        else:
            syn_df = pd.DataFrame(rows)
            for col in c.syn_pt_position_split:
                repopulate_list(col, syn_df)

            sb = generate_statebuilder_post(
                info_cache, c, data_resolution=data_resolution
            )
            try:
                url = make_url_robust(
                    syn_df.sort_values(by=c.num_syn_col, ascending=False),
                    sb,
                    datastack,
                    c,
                )
            except Exception as e:
                return html.Div(str(e))
        return html.A(
            "All Input Link", href=url, target="_blank", style={"font-size": "20px"}
        )

    @app.callback(
        Output("all-output-link", "children"),
        Input("all-output-link-button", "n_clicks"),
        Input("submit-button", "n_clicks"),
        Input("target-table-json", "data"),
        Input("client-info-json", "data"),
        InputDatastack,
        Input("synapse-table-resolution-json", "data"),
        prevent_initial_call=True,
    )
    def generate_all_output_link(_1, _2, rows, info_cache, datastack, data_resolution):
        ctx = callback_context
        if not ctx.triggered:
            return ""
        trigger_src = ctx.triggered[0]["prop_id"].split(".")[0]
        if (
            trigger_src == "submit-button"
            or trigger_src == "client-info-json"
            or trigger_src == "target-table-json"
        ):
            return ""

        if rows is None or len(rows) == 0:
            return html.Div("No outputs to show")
        else:
            syn_df = pd.DataFrame(rows)
            for col in c.syn_pt_position_split:
                repopulate_list(col, syn_df)
            sb = generate_statebuilder_pre(
                info_cache, c, data_resolution=data_resolution
            )

            try:
                url = make_url_robust(
                    syn_df.sort_values(by=c.num_syn_col, ascending=False),
                    sb,
                    datastack,
                    c,
                )
            except Exception as e:
                return html.Div(str(e))
        return html.A(
            "All Output Link", href=url, target="_blank", style={"font-size": "20px"}
        )

    pass


from collections import defaultdict
import standard_transform

transform_lookup = defaultdict(standard_transform.identity_transform)
transform_lookup['minnie65_phase3'] = standard_transform.minnie_transform_nm()
transform_lookup['v1dd'] = standard_transform.v1dd_transform_nm()

def get_transform(aligned_volume):
    return transform_lookup[aligned_volume]

def extract_depth(df, depth_column, position_column, aligned_volume):
    if len(df) == 0:
        df[depth_column] = None
        return df
    tform = get_transform(aligned_volume)
    df[depth_column] = tform.apply_dataframe(position_column, df, projection='y')
    return df

def compute_depth_y(pt, aligned_volume):
    tform = get_transform(aligned_volume)
    return tform.apply_project('y', pt)



from cachetools import cached, TTLCache, keys
SPLIT_SUFFIXES = ["x", "y", "z"]


#json schema column types that can act as potential columns for looking at tables
ALLOW_COLUMN_TYPES = ['integer', 'boolean', 'string', 'float']
# Helper functions for turning schema field names ot column names
def bound_pt_position(pt):
    return f"{pt}_position"

def bound_pt_root_id(pt):
    return f"{pt}_root_id"

def split_pt_position(pt_position):
    return [f"{pt_position}_{suf}" for suf in SPLIT_SUFFIXES]

_schema_cache = TTLCache(maxsize=128, ttl=86_400)
def _schema_key(schema_name, client, **kwargs):
    allow_types = kwargs.get('allow_types', ALLOW_COLUMN_TYPES)
    key = keys.hashkey(schema_name, str(allow_types))
    return key

@cached(cache=_schema_cache, key=_schema_key)
def get_col_info(schema_name, client, spatial_point='BoundSpatialPoint', allow_types=ALLOW_COLUMN_TYPES, omit_fields=[]):
    schema = client.schema.schema_definition(schema_name)
    sp_name = f"#/definitions/{spatial_point}"
    n_sp = 0
    sn = schema['$ref'].split('/')[-1]
    add_cols = []
    for k, v in schema['definitions'][sn]['properties'].items():
        if v.get('$ref','') == sp_name:
            pt_name = k
            n_sp+=1
        else:
            if k in omit_fields:
                continue
            # Field type is format if exists, type otherwise
            if v.get('format', v.get('type')) in allow_types:
                add_cols.append(k)
    if n_sp != 1:
        pt_name = None
    return pt_name, add_cols

_table_cache = TTLCache(maxsize=128, ttl=86_400)
def _table_key(table_name, client, **kwargs):
    merge_schema = kwargs.get('merge_schema', True)
    allow_types = kwargs.get('allow_types', ALLOW_COLUMN_TYPES)
    key = keys.hashkey(table_name, merge_schema, str(allow_types))
    return key

@cached(cache=_table_cache, key=_table_key)
def get_table_info(tn, client, allow_types=ALLOW_COLUMN_TYPES, merge_schema=True):
    """Get the point column and additional columns from a table

    Parameters
    ----------
    tn : str
        Table name
    client : CAVEclient
        Client
    omit_cols : list, optional
        List of strings for tables to omit from the list. By default, ['valid', 'target_id']

    Returns
    -------
    pt
        Point column prefix
    cols
        List of additional columns names
    """
    if tn is None:
        return None, []
    meta = table_metadata(tn, client)
    ref_table = meta.get('reference_table')
    if ref_table is None or merge_schema is False:
        schema = meta['schema']
        extra_cols = []
    else:
        schema = table_metadata(ref_table, client).get('schema')
        _, extra_cols = get_col_info(meta['schema'], client, allow_types=allow_types, omit_fields=['target_id'])
    pt, add_cols = get_col_info(schema, client, allow_types=allow_types)
    cols = add_cols + extra_cols
    return pt, cols

_metadata_cache = TTLCache(maxsize=128, ttl=86_400)
def _metadata_key(tn, client, **kwargs):
    key = keys.hashkey(tn)
    return key

@cached(cache=_metadata_cache, key=_metadata_key)
def table_metadata(table_name, client, meta=None):
    "Caches getting table metadata"
    if meta is None:
        meta = client.materialize.get_table_metadata(table_name)
    if "schema" not in meta:
        meta["schema"] = meta.get('schema_type')
    return meta

_table_list_cache = TTLCache(maxsize=64, ttl=86_400)
def _table_list_key(tables, client):
    key = keys.hashkey('_'.join(tables))
    return key

@cached(cache=_table_list_cache, key=_table_list_key)
def populate_metadata_cache(tables, client):
    all_meta = client.materialize.get_tables_metadata()
    for tn, meta in zip(tables, all_meta):
        table_metadata(tn, client, meta=meta)
    pass



from ..common.config import CommonConfig


class ConnectivityConfig(CommonConfig):
    def __init__(self, config):
        super().__init__(config)

        self.synapse_aggregation_rules = config.get("synapse_aggregation_rules", {})
        self.aggregation_columns = list(self.synapse_aggregation_rules.keys())
        self.table_columns = (
            [
                self.root_id_col,
                self.num_syn_col,
            ]
            + self.aggregation_columns
            + [self.num_soma_col]
        )



import copy
###########################################
### Default data and request parameters ###
###########################################
from .schema_utils import bound_pt_position, split_pt_position, bound_pt_root_id

DATA_RESOLUTION = [1,1,1]

def parse_environ_vector(input, num_type):
    return [num_type(x) for x in input.split(",")]


class CommonConfig(object):
    def __init__(self, config):
        self.default_datastack = config.get("datastack")
        if self.default_datastack is None:
            raise ValueError("Must datastack parameter!")

        self.server_address = config.get("server_address")
        if self.server_address is None:
            raise ValueError("Must set server address parameter!")

        self.disallow_live_query = config.get("disallow_live_query", False)
        self.image_black = config.get("image_black", 0)
        self.image_white = config.get("image_white", 1)

        self.target_root_id_per_call = config.get("target_root_id_per_call", 200)
        self.max_chunks = config.get("max_chunks", 20)
        self.pool_maxsize = 2 * self.max_chunks
        self.voxel_resolution = config.get("voxel_resolution")

        self.data_resolution = DATA_RESOLUTION

        ##############################
        ### Link generation limits ###
        ##############################

        self.max_dataframe_length = config.get("max_dataframe_length", 8_000)
        self.max_server_dataframe_length = config.get(
            "max_server_dataframe_length", 20_000
        )

        # If None, the info service is used
        self.nucleus_table = config.get("nucleus_table", None)
        self.nucleus_id_column = config.get("nucleus_id_column", "id")
        self.nucleus_filter = config.get('nucleus_filter', {})

        # Used to look up number of neurons per root id
        self.soma_table = self.nucleus_table
        self.soma_id_column = self.nucleus_id_column

        # Used to look up connectivity
        # If None, the info service is used
        self.synapse_table = config.get("synapse_table", None)
        self.syn_id_col = "id"
        self.pre_pt_root_id = "pre_pt_root_id"
        self.post_pt_root_id = "post_pt_root_id"
        self.synapse_aggregation_rules = config.get("synapse_aggregation_rules", {})

        self.ct_cell_type_point = None

        self.syn_pt_prefix = config.get("syn_position_column", "ctr_pt")
        self.syn_pt_position = bound_pt_position(self.syn_pt_prefix)
        self.syn_pt_position_split = split_pt_position(self.syn_pt_position)

        self.soma_pt_prefix = config.get("soma_postion_column", "pt")
        self.soma_pt_position = bound_pt_position(self.soma_pt_prefix)
        self.soma_pt_root_id = bound_pt_root_id(self.soma_pt_prefix)

        self.soma_ct_col = config.get("soma_cell_type_column", "cell_type")
        self.soma_table_cell_category = config.get("soma_table_cell_type")
        if self.soma_ct_col and self.soma_table_cell_category:
            self.soma_table_query = (
                f"{self.soma_ct_col} == '{self.soma_table_cell_category}'"
            )
        else:
            self.soma_table_query = None

        self.num_soma_prefix = "num"
        self.num_syn_col = "num_syn"
        self.root_id_col = "root_id"
        self.num_soma_suffix = "_soma"
        self.num_soma_col = f"{self.num_soma_prefix}{self.num_soma_suffix}"

        self.soma_position_agg = self.soma_pt_position + self.num_soma_suffix

        self.synapse_table_columns_base = [
            "id",
            self.pre_pt_root_id,
            self.post_pt_root_id,
        ] + split_pt_position(self.syn_pt_position)

        additional_syn_merges = []
        for _, v in self.synapse_aggregation_rules.items():
            if v["column"] not in additional_syn_merges:
                additional_syn_merges.append(v["column"])

        self.synapse_table_columns_dataframe = (
            self.synapse_table_columns_base + additional_syn_merges
        )

        self.target_table_display = (
            [self.root_id_col]
            + split_pt_position(self.syn_pt_position)
            + [
                self.num_syn_col,
                self.num_soma_col,
            ]
            + list(self.synapse_aggregation_rules.keys())
        )

        self.soma_table_columns = [
            self.soma_pt_root_id,
            self.num_soma_col,
        ] + split_pt_position(self.soma_pt_position)

    @property
    def ct_cell_type_pt_position(self):
        if self.ct_cell_type_point is None:
            return None
        return bound_pt_position(self.ct_cell_type_point)

    @property
    def ct_cell_type_pt_position_split(self):
        if self.ct_cell_type_point is None:
            return None
        return split_pt_position(self.ct_cell_type_pt_position)

    @property
    def ct_cell_type_root_id(self):
        if self.ct_cell_type_point is None:
            return None
        return bound_pt_root_id(self.ct_cell_type_point)


def RegisterTable(pt, value_columns, config):
    config = copy.deepcopy(config)
    config.ct_cell_type_point = pt
    config.value_columns = [config.nucleus_id_column] + value_columns
    return config


from plotly import colors
import plotly.graph_objects as go
import numpy as np
import pandas as pd

def _violin_plot(syn_df, x_col, y_col, name, side, color, xaxis, yaxis):
    return go.Violin(
        x=syn_df[x_col],
        y=syn_df[y_col],
        side=side,
        scalegroup="syn",
        spanmode='hard',
        name=name,
        points=False,
        line_color=f"rgb{color}",
        fillcolor=f"rgb{color}",
        xaxis=xaxis,
        yaxis=yaxis,
        hoverinfo='text',
        hovertext=f"{len(syn_df)} Syn.",
        # bandwidth=0.2,
    )


def post_violin_plot(
    ndat,
    xaxis=None,
    yaxis=None,
):
    return _violin_plot(
        ndat.syn_all_df().query('direction == "post"'),
        x_col="x",
        y_col=ndat.config.synapse_depth_column,
        name="Post",
        side="negative",
        color=ndat.config.vis.dendrite_color,
        xaxis=xaxis,
        yaxis=yaxis,
    )


def pre_violin_plot(
    ndat,
    xaxis=None,
    yaxis=None,
):
    return _violin_plot(
        ndat.syn_all_df().query('direction == "pre"'),
        x_col="x",
        y_col=ndat.config.synapse_depth_column,
        name="Pre",
        side="positive",
        color=ndat.config.vis.axon_color,
        xaxis=xaxis,
        yaxis=yaxis,
    )

from itertools import cycle
def _colorscheme(n):
    if n <= 10:
        clrs = cycle(colors.qualitative.G10)
    else:
        clrs = cycle(colors.qualitative.Dark24)
    return [next(clrs) for i in range(n)]

def synapse_soma_scatterplot(
    targ_df,
    config,
    color_column,
    xaxis=None,
    yaxis=None,
):

    if color_column is None or color_column == "":
        fake_cell_type_column = 'DummyColumn_'
        while fake_cell_type_column in targ_df.columns:
            fake_cell_type_column += 'a'
        color_column = fake_cell_type_column
        targ_df[color_column] = config.null_cell_type_label
        ctypes = [config.null_cell_type_label]
    else:
        if targ_df[color_column].dtype == 'float64':
            targ_df[color_column] = targ_df[color_column].astype(pd.Int64Dtype())
            ctypes = sorted(list(np.unique(targ_df[color_column].dropna()).astype(str)))
            targ_df[color_column] = targ_df[color_column].astype(str).replace(
                {'<NA>': config.null_cell_type_label}
            )
        else:
            ctypes = sorted(list(np.unique(targ_df[color_column].dropna()).astype(str)))
            targ_df[color_column] = targ_df[color_column].fillna(config.null_cell_type_label).astype(str)
        ctypes = ctypes+[config.null_cell_type_label]

    
    if len(ctypes)>1:
        cmap = _colorscheme(len(ctypes)-1) + ['#333333']
    else:
        cmap = ['#333333']

    alpha_default = {config.null_cell_type_label: 0.2}
    panels = []
    alpha= config.vis.e_opacity
    for ct, clr in zip(ctypes, cmap):
        targ_df_r = targ_df.query(f"{color_column}=='{ct}'")
        panel = go.Scattergl(
            x=targ_df_r[config.soma_depth_column],
            y=targ_df_r[config.synapse_depth_column],
            mode="markers",
            marker=dict(
                color=clr,
                line_width=0,
                size=4,
                opacity=alpha_default.get(ct, alpha),
            ),
            xaxis=xaxis,
            yaxis=yaxis,
            name=ct,
            # hoverinfo='none',
        )
        panels.append(panel)

    return panels

def bar_plot_df(
    targ_df,
    config,
    color_value,
):

    targ_df = targ_df.replace({config.null_cell_type_label: None}).dropna(subset=color_value)
    xtypes = sorted(list(np.unique(targ_df[color_value])))
    clrs = _colorscheme(len(xtypes))
    cnts = targ_df.value_counts(color_value).loc[xtypes]

    bar = go.Bar(
        name=color_value,
        x=xtypes,
        y=cnts,
        text=cnts,
        marker_color=clrs,
    )
    return bar


def bar_data(
    ndat,
    cell_type_column,
    num_syn_column,
):
    targ_df = ndat.partners_out().dropna(subset=[cell_type_column])
    return targ_df.groupby(cell_type_column)[num_syn_column].sum()


def _bar_plot(
    bar_data,
    name,
    color,
):
    return go.Bar(
        name=name,
        x=bar_data.values,
        y=bar_data.index,
        marker_color=f"rgb{color}",
        orientation="h",
    )


def _format_color(color, alpha=None):
    color = tuple(np.floor(255 * np.array(color)).astype(int))
    if alpha is None:
        return color
    else:
        return tuple(list(color) + [alpha])


def _prepare_bar_plot(
    ndat,
    cell_type_column,
    color,
    cell_types,
    valence,
):
    if valence == "u":
        if cell_types is None:
            cell_types = np.unique(
                ndat.property_data(ndat.cell_type_table)[cell_type_column]
            )
        name = "Targets"
    else:
        if valence == "i":
            map_ind = "i"
            name = "I Targets"
        elif valence == "e":
            map_ind = "e"
            name = "E Targets"

        if cell_types is None:
            cell_types = (
                ndat.property_data(ndat.cell_type_table)
                .groupby(ndat.valence_map["column"])
                .agg({cell_type_column: np.unique})
                .loc[ndat.valence_map[map_ind]][cell_type_column]
            )

    bdat = bar_data(ndat, cell_type_column, ndat.config.num_syn_col)

    # Fill in any cell types in the table
    for ct in cell_types:
        if ct not in bdat.index:
            bdat.loc[ct] = 0

    return _bar_plot(
        bdat.sort_index().loc[cell_types],
        name,
        _format_color(color),
    )


def excitatory_bar_plot(
    ndat,
    cell_type_column,
    cell_types=None,
):
    return _prepare_bar_plot(
        ndat, cell_type_column, ndat.config.vis.e_color, cell_types, "e"
    )


def inhibitory_bar_plot(
    ndat,
    cell_type_column,
    cell_types=None,
):
    return _prepare_bar_plot(
        ndat, cell_type_column, ndat.config.vis.i_color, cell_types, "i"
    )


def uniform_bar_plot(
    ndat,
    cell_type_column,
    cell_types=None,
):
    return _prepare_bar_plot(
        ndat, cell_type_column, ndat.config.vis.u_color, cell_types, "u"
    )


import numpy as np
import seaborn as sns
import pandas as pd
import os
import pathlib
from ..common.config import CommonConfig
from ..common.schema_utils import bound_pt_position, bound_pt_root_id

####################
### Column names ###
####################


class VisConfig:
    def __init__(
        self,
        dendrite_color,
        axon_color,
        e_palette,
        i_palette,
        u_palette,
        tick_locs=[],
        tick_labels=[],
        base_ind=6,
        n_e_colors=9,
        n_i_colors=9,
        n_u_colors=9,
        e_string="Exc",
        i_string="Inh",
        u_string="Unknown",
        e_opacity=0.5,
        i_opacity=0.75,
        u_opacity=0.3,
    ):
        self.dendrite_color = dendrite_color
        self.axon_color = axon_color

        self.e_colors = sns.color_palette(e_palette, n_colors=n_e_colors)
        self.i_colors = sns.color_palette(i_palette, n_colors=n_i_colors)
        self.u_colors = sns.color_palette(u_palette, n_colors=n_u_colors)
        self.base_ind = base_ind

        self.e_string = e_string
        self.i_string = i_string
        self.u_string = u_string

        self.e_opacity = e_opacity
        self.i_opacity = i_opacity
        self.u_opacity = u_opacity

        self.ticklocs = tick_locs
        self.tick_labels = tick_labels

    @property
    def clrs(self):
        return np.array([self.axon_color, self.dendrite_color])

    @property
    def e_color(self):
        return self.e_colors[self.base_ind]

    @property
    def i_color(self):
        return self.i_colors[self.base_ind]

    @property
    def u_color(self):
        return self.u_colors[max(self.base_ind - 2, 0)]

    @property
    def valence_colors(self):
        return np.vstack([self.e_color, self.i_color, self.u_color])

    def valence_color_map(self, is_inhib):
        cmap = []
        for x in is_inhib:
            if pd.isna(x):
                cmap.append(2)
            elif x:
                cmap.append(0)
            else:
                cmap.append(1)
        return np.array(cmap)

    def valence_string_map(self, is_inhib):
        smap = []
        for x in is_inhib:
            if pd.isna(x):
                smap.append(self.u_string)
            elif x:
                smap.append(self.i_string)
            else:
                smap.append(self.e_string)
        return smap


class TypedConnectivityConfig(CommonConfig):
    def __init__(self, config):
        super().__init__(config)

        self.ct_conn_cell_type_column = config.get(
            "ct_conn_cell_type_column", "cell_type"
        )

        self.soma_depth_column = config.get("ct_conn_soma_depth_column", "soma_depth")
        self.synapse_depth_column = config.get("ct_conn_syn_depth_column", "syn_depth")
        self.is_inhibitory_column = config.get(
            "ct_conn_is_inhibitory_column", "is_inhibitory"
        )

        self.cell_type_dropdown_options = config.get("cell_type_dropdown_options", [])
        self.default_cell_type_option = config.get("default_cell_type_option", "")

        self.omit_cell_type_tables = config.get("omit_cell_type_tables", [])
        self.synapse_aggregation_rules = config.get("synapse_aggregation_rules", {})
        self.aggregation_columns = list(self.synapse_aggregation_rules.keys())
        self.table_valence_map = config.get("valence_map", {})

        self.table_columns = (
            [
                self.root_id_col,
                self.num_syn_col,
                self.soma_depth_column,
            ]
            + self.aggregation_columns
            + [self.num_soma_col]
        )

        self.show_plots = config.get("ct_conn_show_plots", True)
        self.show_depth_plots = config.get("ct_conn_show_depth_plots", True)
        self.null_cell_type_label = config.get('null_cell_type_label', "No Type")

        # self.layer_bnds = np.load(f"{data_path}/layer_bounds_v1.npy")
        # self.height_bnds = np.load(f"{data_path}/height_bounds_v1.npy")
        self.height_bnds = config.get('height_bounds', [])
        if self.height_bnds is not None:
            self.height_bnds = np.array(self.height_bnds)
        self.layer_bnds = config.get('layer_bounds', [])
        if self.layer_bnds is not None:
            self.layer_bnds = np.array(self.layer_bnds)
        self.layer_labels = config.get('layer_labels', [])

        if self.layer_bnds is not None and self.height_bnds is not None:
            ticklocs = np.concatenate(
                [self.height_bnds[0:1], self.layer_bnds, self.height_bnds[1:]]
            )
        else:
            ticklocs = np.array([])

        if self.layer_labels is None:
            self.layer_labels = self.layer_bnds.astype(str)


        dendrite_color = config.get("ct_conn_dendrite_color", (0.894, 0.102, 0.110))
        axon_color = config.get("ct_conn_axon_color", (0.227, 0.459, 0.718))

        e_color_palette = config.get("ct_conn_e_palette", "RdPu")
        i_color_palette = config.get("ct_conn_i_palette", "Greens")
        u_color_palette = config.get("ct_conn_u_palette", "Greys")
        base_ind = int(config.get("ct_conn_palette_base", 6))

        self.vis = VisConfig(
            dendrite_color,
            axon_color,
            e_color_palette,
            i_color_palette,
            u_color_palette,
            base_ind=base_ind,
            tick_locs=ticklocs,
            tick_labels=["L1", "L2/3", "L4", "L5", "L6", "WM", ""],
        )




import dash_bootstrap_components as dbc

# See https://dash-bootstrap-components.opensource.faculty.ai/docs/
# for using dash_bootstrap_components for simplifying programmatic layout.
# Set this to an empty list if you are not using any external sheets
external_stylesheets = [dbc.themes.YETI]



from dash import Dash

from .callbacks import register_callbacks
from .layout import title, page_layout, app_layout
from ..common.external_stylesheets import external_stylesheets
from ..common.dash_url_helper import setup
import flask


def create_app(name=__name__, config={}, **kwargs):
    if "external_stylesheets" not in kwargs:
        kwargs["external_stylesheets"] = external_stylesheets
    app = Dash(name, **kwargs)
    app.title = title
    app.layout = app_layout
    setup(app, page_layout=page_layout)
    register_callbacks(app, config)
    return app



import datetime
import pandas as pd
from functools import partial

from dash import dcc, html, callback_context
from dash.dependencies import Input, Output, State
import dash_bootstrap_components as dbc

from .config import TypedConnectivityConfig
from ..common import table_lookup as tbl
from ..common.link_utilities import (
    generate_statebuilder,
    generate_statebuilder_pre,
    generate_statebuilder_post,
    generate_statebuider_syn_grouped,
    generate_statebuilder_syn_cell_types,
    EMPTY_INFO_CACHE,
    MAX_URL_LENGTH,
    make_url_robust,
    aligned_volume,
)
from ..common.dash_url_helper import _COMPONENT_ID_TYPE
from ..common.lookup_utilities import (
    get_type_tables,
    make_client,
)
from ..common.schema_utils import get_table_info
from ..common.dataframe_utilities import (
    stringify_root_ids, stringify_list, rehydrate_dataframe, rebuild_synapse_dataframe
)
from ..cell_type_table.config import CellTypeConfig
from .neuron_data_cortex import NeuronDataCortex as NeuronData
from .neuron_data_cortex import ALLOW_COLUMN_TYPES_DISCRETE
from .cortex_panels import *

try:
    from loguru import logger
    import time
except:
    logger = None


InputDatastack = Input({"id_inner": "datastack", "type": _COMPONENT_ID_TYPE}, "value")
OutputDatastack = Output({"id_inner": "datastack", "type": _COMPONENT_ID_TYPE}, "value")

StateRootID = State({"id_inner": "anno-id", "type": _COMPONENT_ID_TYPE}, "value")
StateCellTypeTable = State(
    {"id_inner": "cell-type-table-dropdown", "type": _COMPONENT_ID_TYPE},
    "value",
)
OutputCellTypeMenuOptions = Output(
    {"id_inner": "cell-type-table-dropdown", "type": _COMPONENT_ID_TYPE},
    "options",
)
OutputCellTypeValue = Output(
    {"id_inner": "cell-type-table-dropdown", "type": _COMPONENT_ID_TYPE},
    "value",
)

StateAnnoType = State({"id_inner": "id-type", "type": _COMPONENT_ID_TYPE}, "value")
StateLiveQuery = State(
    {"id_inner": "live-query-toggle", "type": _COMPONENT_ID_TYPE}, "value"
)
StateLinkGroupValue = State('group-by', 'value')

OutputLiveQueryToggle = Output(
    {"id_inner": "live-query-toggle", "type": _COMPONENT_ID_TYPE},
    "options",
)
OutputLiveQueryValue = Output(
    {"id_inner": "live-query-toggle", "type": _COMPONENT_ID_TYPE}, "value"
)




def allowed_action_trigger(ctx, allowed_buttons):
    if not ctx.triggered:
        return False
    trigger_src = ctx.triggered[0]["prop_id"].split(".")[0]
    return trigger_src in allowed_buttons


def generic_syn_link_generation(
    sb_function,
    rows,
    info_cache,
    datastack,
    config,
    link_text,
    item_name="synapses",
):
    if rows is None or len(rows) == 0:
        return html.Div(f"No {item_name} to show")
    else:
        syn_df = rehydrate_dataframe(rows, config.syn_pt_position_split)
        sb = sb_function(info_cache)
    try:
        url = make_url_robust(
            syn_df.sort_values(by=config.num_syn_col, ascending=False),
            sb,
            datastack,
            config,
        )
    except Exception as e:
        return html.Div(str(e))

    return html.A(link_text, href=url, target="_blank", style={"font-size": "20px"})


def make_violin_plot(ndat, height=350):
    if ndat is None:
        return html.Div("")

    violin = violin_fig(ndat, height=height)
    contents = [
        html.H5("Input/Output Depth", className='card-title'),
        dcc.Graph(
            figure=violin, style={"margin-left": "5rem", "margin-right": "5rem", "width": "auto", "height": "25rem"}
        ),
    ]
    return contents


def make_ct_plots(rows, config, aligned_volume, color_column):
    if rows is None:
        return html.Div(""), html.Div("")
    if len(rows) == 0:
        return html.Div(""), html.Div("")

    df = rebuild_synapse_dataframe(
        rows,
        config,
        aligned_volume,
        value_cols=[color_column],
    )

    if color_column == "":
        color_column = None

    if config.show_depth_plots:
        scatter_contents = make_scatter_div(df.copy(), config, color_column, width=450, height=350)
    else:
        scatter_contents = html.Div("")
    bar_contents = make_bar_div(df.copy(), config, color_column, width = 450, height=350)

    return scatter_contents, bar_contents


def make_scatter_div(
    df, config, color_column, width=450, height=350,
):
    scatter_fig = scatter_fig_df(df, config, color_column, width, height)
    contents = [
        html.H5("Synapse/Target Soma Depth", className='card-title'),
        dcc.Graph(
            figure=scatter_fig, style={"height": "100%",  "width": "auto"},
        ),
    ]
    return contents

def make_bar_div(df, config, color_column, width=450, height=350):
    if color_column is None:
        contents = [
            html.H4(
                "Select value column for bar chart",
                style=
                    {
                        'text-align': 'center', 'vertical-align': 'middle'
                    },
                ),
        ]
    else:
        bar_fig = bar_fig_df(df, config, color_column)
        contents = [
            html.H5("Target Distribution", style={'text-align': 'center'}),
            dcc.Graph(
                figure=bar_fig, style={"width": "auto", "height": "100%"}
            )
        ]
    return contents

def register_callbacks(app, config):

    c = TypedConnectivityConfig(config)

    @app.callback(
        Output("data-table", "selected_rows"),
        Input("reset-selection", "n_clicks"),
        Input("connectivity-tab", "value"),
    )
    def reset_selection(n_clicks, tab_value):
        return []

    @app.callback(
        Output("header-bar", 'children'),
        InputDatastack,
    )
    def set_header(datastack):
        return html.H3(f"Typed Connectivity Viewer — {datastack}", className="bg-primary text-white p-2 mb-2 text-center")


    @app.callback(
        Output("data-table", "columns"),
        Output("group-by", "options"),
        Input("submit-button", "n_clicks"),
        InputDatastack,
        StateCellTypeTable,
    )
    def define_table_columns(_, datastack, cell_type_table):
        client = make_client(datastack, c.server_address)
        if cell_type_table == "" or cell_type_table is None:
            return [{"name": i, "id": i} for i in c.table_columns], []

        _, val_cols = get_table_info(cell_type_table, client, allow_types=ALLOW_COLUMN_TYPES_DISCRETE)
        table_cons = c.table_columns + val_cols
        return (
            [{"name": i, "id": i} for i in table_cons],
            [{"label": k, "value": k} for k in val_cols],
        )

    @app.callback(
        Output('plot-color-value', 'options'),
        Output('plot-color-value', 'value'),
        Input('group-by', 'options'),
        Input('plot-color-value', 'value'),
    )
    def update_plot_options(new_options, old_value):
        DEFAULT_VALUE = 'cell_type'
        option_labels = [v['label'] for v in new_options]
        if old_value in option_labels:
            new_value = old_value
        elif DEFAULT_VALUE in new_options:
            new_value = DEFAULT_VALUE
        else:
            new_value = ''
        return new_options, new_value

    @app.callback(
        OutputDatastack,
        InputDatastack,
    )
    def define_datastack(datastack):
        if datastack is None:
            datastack = ""

        if len(datastack) == 0:
            return c.default_datastack
        else:
            return datastack

    @app.callback(
        OutputLiveQueryToggle,
        OutputLiveQueryValue,
        InputDatastack,
        StateLiveQuery,
    )
    def disable_live_query(_, lq):
        options_active = [{"label": "Live Query", "value": 1}]
        options_disabled = [{"label": "Live Query", "value": 1, "disabled": True}]
        if c.disallow_live_query:
            return options_disabled, ""
        else:
            return options_active, lq

    @app.callback(
        OutputCellTypeMenuOptions,
        InputDatastack,
    )
    def set_cell_type_dropdown(datastack):
        return get_type_tables(datastack, c)

    @app.callback(
        OutputCellTypeValue,
        InputDatastack,
        StateCellTypeTable,
    )
    def default_cell_type_option(_, curr_value):
        if curr_value != "":
            return curr_value
        else:
            return c.default_cell_type_option

    @app.callback(
        Output("message-text", "children"),
        Output("message-text", "color"),
        Output("main-loading-placeholder", "children"),
        Output("target-table-json", "data"),
        Output("source-table-json", "data"),
        Output("output-tab", "label"),
        Output("input-tab", "label"),
        Output("reset-selection", "n_clicks"),
        Output("client-info-json", "data"),
        Output("violin-plot", "children"),
        Output("synapse-table-resolution-json", "data"),
        Input("submit-button", "n_clicks"),
        InputDatastack,
        StateRootID,
        StateAnnoType,
        StateCellTypeTable,
        StateLiveQuery,
    )
    def update_data(_1, datastack_name, anno_id, id_type, ct_table_value, query_toggle):
        if logger is not None:
            t0 = time.time()

        try:
            client = make_client(datastack_name, c.server_address)
            info_cache = client.info.info_cache[datastack_name]
            info_cache["global_server"] = client.server_address
        except Exception as e:
            return (
                html.Div(str(e)),
                "danger",
                "",
                [],
                [],
                "Output",
                "Input",
                1,
                EMPTY_INFO_CACHE,
                make_violin_plot(None),
                None,
            )

        if not ct_table_value:
            ct_table_value = None
        info_cache["cell_type_column"] = ct_table_value

        if len(query_toggle) == 1 and not c.disallow_live_query:
            live_query = True
        else:
            live_query = False

        if live_query:
            timestamp = datetime.datetime.utcnow()
            timestamp_ngl = None
            info_cache['ngl_timestamp'] = None
        else:
            timestamp = client.materialize.get_timestamp()
            timestamp_ngl = timestamp
            info_cache["ngl_timestamp"] = timestamp.timestamp()

        if anno_id is None or len(anno_id) == 0:
            return (
                html.Div("Please select a cell id and press Submit"),
                "info",
                "",
                [],
                [],
                "Output",
                "Input",
                1,
                EMPTY_INFO_CACHE,
                make_violin_plot(None),
                None,
            )
        else:
            if id_type == "root_id":
                object_id = int(anno_id)
                object_id_type = "root"
            elif id_type == "nucleus_id":
                object_id = int(anno_id)
                object_id_type = "nucleus"
            else:
                raise ValueError('id_type must be either "root_id" or "nucleus_id"')
        
        try:
            nrn_data = NeuronData(
                object_id=object_id,
                client=client,
                config=c,
                value_table=ct_table_value,
                timestamp=timestamp,
                id_type=object_id_type,
                is_live=live_query,
                n_threads=2,
            )

            root_id = nrn_data.root_id
            info_cache["root_id"] = str(root_id)

            pre_targ_df = nrn_data.partners_out_plus()
            pre_targ_df = stringify_root_ids(
                pre_targ_df, stringify_cols=[c.root_id_col]
            )

            post_targ_df = nrn_data.partners_in_plus()
            post_targ_df = stringify_root_ids(
                post_targ_df, stringify_cols=[c.root_id_col]
            )

            n_syn_pre = pre_targ_df[c.num_syn_col].sum()
            n_syn_post = post_targ_df[c.num_syn_col].sum()

            for col in nrn_data.config.syn_pt_position_split:
                stringify_list(col, pre_targ_df)
                stringify_list(col, post_targ_df)


            if logger is not None:
                logger.info(
                    f"Data update for {root_id} | time:{time.time() - t0:.2f} s, syn_in: {len(pre_targ_df)} , syn_out: {len(post_targ_df)}"
                )
            if nrn_data.nucleus_id is not None and nrn_data.soma_table is not None:
                if np.issubdtype(type(nrn_data.nucleus_id), np.integer):
                    nuc_id_text = f"  (nucleus id: {nrn_data.nucleus_id})"
                else:
                    nuc_id_text = f" (Multiple nucleus ids in segment: {', '.join([str(x) for x in nrn_data.nucleus_id])})"
            else:
                nuc_id_text = ""
            if ct_table_value:
                ct_text = f"table {ct_table_value}"
            else:
                ct_text = "no cell type table"

            if live_query:
                message_text = f"Current connectivity for root id {root_id}{nuc_id_text} and {ct_text}"
            else:
                message_text = f"Connectivity for root id {root_id}{nuc_id_text} and {ct_text} materialized on {timestamp_ngl:%m/%d/%Y} (v{client.materialize.version})"

            if c.show_depth_plots:
                vplot = make_violin_plot(nrn_data, None)
            else:
                vplot = make_violin_plot(None)
            syn_res = nrn_data.synapse_data_resolution
            del nrn_data
            del client

            return (
                html.Div(message_text),
                "success",
                "",
                pre_targ_df.to_dict("records"),
                post_targ_df.to_dict("records"),
                f"Output (n = {n_syn_pre})",
                f"Input (n = {n_syn_post})",
                1,
                info_cache,
                vplot,
                syn_res,
            )
        except Exception as e:
            return (
                html.Div(str(e)),
                "danger",
                "",
                [],
                [],
                "Output",
                "Input",
                1,
                EMPTY_INFO_CACHE,
                make_violin_plot(None),
                None,
            )

    @app.callback(
        Output('scatter-plot', 'children'),
        Output('bar-plot', 'children'),
        Input("target-table-json", "data"),
        Input("plot-color-value", 'value'),
        Input("client-info-json", "data"),
    )
    def update_scatter_bar_plots(rows, color_column, info_cache):
        return make_ct_plots(
            rows,
            c,
            aligned_volume(info_cache),
            color_column,
        )

    @app.callback(
        Output("data-table", "data"),
        Input("connectivity-tab", "value"),
        Input("target-table-json", "data"),
        Input("source-table-json", "data"),
    )
    def update_table(
        tab_value,
        pre_data,
        post_data,
    ):
        if tab_value == "tab-pre":
            return pre_data
        elif tab_value == "tab-post":
            return post_data
        else:
            return []

    @app.callback(
        Output("ngl-link", "href"),
        Output("ngl-link", "children"),
        Output("ngl-link", "disabled"),
        Output("link-loading", "children"),
        Input("connectivity-tab", "value"),
        Input("data-table", "derived_virtual_data"),
        Input("data-table", "derived_virtual_selected_rows"),
        Input("client-info-json", "data"),
        Input("synapse-table-resolution-json", "data"),
    )
    def update_link(
        tab_value,
        rows,
        selected_rows,
        info_cache,
        synapse_data_resolution,
    ):
        large_state_text = (
            "Table Too Large - Please Filter or Use Whole Cell Neuroglancer Links"
        )

        def small_state_text(n):
            return f"Neuroglancer: ({n} partners)"

        if info_cache is None:
            return "", "No datastack set", True, ""

        if rows is None or len(rows) == 0:
            rows = {}
            sb = generate_statebuilder(info_cache, c)
            return (
                sb.render_state(None, return_as="url"),
                small_state_text(0),
                False,
                "",
            )
        else:
            syn_df = rehydrate_dataframe(rows, c.syn_pt_position_split)

            if len(selected_rows) == 0:
                if tab_value == "tab-pre":
                    sb = generate_statebuilder_pre(
                        info_cache, c, data_resolution=synapse_data_resolution
                    )
                elif tab_value == "tab-post":
                    sb = generate_statebuilder_post(
                        info_cache, c, data_resolution=synapse_data_resolution
                    )
                else:
                    raise ValueError('tab must be "tab-pre" or "tab-post"')
                url = sb.render_state(
                    syn_df.sort_values(by=c.num_syn_col, ascending=False),
                    return_as="url",
                )
                small_out_text = small_state_text(len(syn_df))

            else:
                if tab_value == "tab-pre":
                    anno_layer = "Output Synapses"
                elif tab_value == "tab-post":
                    anno_layer = "Input Synapses"
                sb = generate_statebuider_syn_grouped(
                    info_cache,
                    anno_layer,
                    c,
                    preselect=len(selected_rows) == 1,
                    data_resolution=synapse_data_resolution,
                )
                url = sb.render_state(
                    syn_df.iloc[selected_rows].sort_values(
                        by=c.num_syn_col, ascending=False
                    ),
                    return_as="url",
                )
                small_out_text = small_state_text(len(selected_rows))

        if len(url) > MAX_URL_LENGTH:
            return "", large_state_text, True, ""
        else:
            return url, small_out_text, False, ""

    @app.callback(
        Output("all-input-link", "children"),
        Output("all-input-link-button", "children"),
        Output("all-input-link-button", "disabled"),
        Input("all-input-link-button", "n_clicks"),
        Input("all-input-link-button", "children"),
        Input("submit-button", "n_clicks"),
        Input("source-table-json", "data"),
        Input("client-info-json", "data"),
        InputDatastack,
        Input("synapse-table-resolution-json", "data"),
        prevent_initial_call=True,
    )
    def generate_all_input_link(
        _1, _2, curr, rows, info_cache, datastack, data_resolution
    ):
        if not allowed_action_trigger(callback_context, ["all-input-link-button"]):
            return "  ", "Generate Link", False
        return (
            generic_syn_link_generation(
                partial(
                    generate_statebuilder_post,
                    config=c,
                    data_resolution=data_resolution,
                ),
                rows,
                info_cache,
                datastack,
                c,
                "Neuroglancer Link",
                "Inputs",
            ),
            "Link Generated",
            True,
        )

    @app.callback(
        Output("cell-typed-input-link", "children"),
        Output("cell-typed-input-link-button", "children"),
        Output("cell-typed-input-link-button", "disabled"),
        Input("cell-typed-input-link-button", "n_clicks"),
        Input("submit-button", "n_clicks"),
        Input("source-table-json", "data"),
        Input("client-info-json", "data"),
        InputDatastack,
        Input("synapse-table-resolution-json", "data"),
        Input('group-by', 'value'),
        Input('no-type-annotation', 'value'),
        prevent_initial_call=True,
    )
    def generate_cell_typed_input_link(
        _1, _2, rows, info_cache, datastack, data_resolution, value_column, include_no_type,
    ):
        if value_column is None or value_column == "":
            return "  ", "No Annotation Column Set", True
        if not allowed_action_trigger(
            callback_context, ["cell-typed-input-link-button"]
        ):
            return "  ", "Generate Link", False

        include_no_type = 1 in include_no_type

        sb, dfs = generate_statebuilder_syn_cell_types(
            info_cache,
            rows,
            c,
            cell_type_column=value_column,
            multipoint=True,
            fill_null="NoType",
            data_resolution=data_resolution,
            include_no_type=include_no_type,
        )
        try:
            url = make_url_robust(dfs, sb, datastack, c)
        except Exception as e:
            return html.Div(str(e))
        return (
            html.A(
                "Grouped Input Link",
                href=url,
                target="_blank",
                style={"font-size": "20px"},
            ),
            "Link Generated",
            True,
        )

    @app.callback(
        Output("all-output-link", "children"),
        Output("all-output-link-button", "children"),
        Output("all-output-link-button", "disabled"),
        Input("all-output-link-button", "n_clicks"),
        Input("submit-button", "n_clicks"),
        Input("target-table-json", "data"),
        Input("client-info-json", "data"),
        InputDatastack,
        Input("synapse-table-resolution-json", "data"),
        prevent_initial_call=True,
    )
    def generate_all_output_link(_1, _2, rows, info_cache, datastack, data_resolution):
        if not allowed_action_trigger(callback_context, ["all-output-link-button"]):
            return "", "Generate Link", False
        return (
            generic_syn_link_generation(
                partial(
                    generate_statebuilder_pre, config=c, data_resolution=data_resolution
                ),
                rows,
                info_cache,
                datastack,
                c,
                "All Output Link",
                "Outputs",
            ),
            "Link Generated",
            True,
        )

    @app.callback(
        Output("cell-typed-output-link", "children"),
        Output("cell-typed-output-link-button", "children"),
        Output("cell-typed-output-link-button", "disabled"),
        Input("cell-typed-output-link-button", "n_clicks"),
        Input("submit-button", "n_clicks"),
        Input("target-table-json", "data"),
        Input("client-info-json", "data"),
        InputDatastack,
        Input("synapse-table-resolution-json", "data"),
        Input('group-by', 'value'),
        Input('no-type-annotation', 'value'),
        prevent_initial_call=True,
    )
    def generate_cell_typed_output_link(
        _1, _2, rows, info_cache, datastack, data_resolution, value_column, include_no_type,
    ):
        if value_column is None or value_column == "":
            return "  ", "No Annotation Column Set", True
        
        if not allowed_action_trigger(
            callback_context, ["cell-typed-output-link-button"]
        ):
            return "  ", "Generate Link", False

        include_no_type = 1 in include_no_type

        sb, df_dict = generate_statebuilder_syn_cell_types(
            info_cache,
            rows,
            c,
            cell_type_column=value_column,
            multipoint=True,
            fill_null="NoType",
            data_resolution=data_resolution,
            include_no_type=include_no_type,
        )
        try:
            url = make_url_robust(df_dict, sb, datastack, c)
        except Exception as e:
            return html.Div(str(e))
        return (
            html.A(
                "Cell Typed Output Link",
                href=url,
                target="_blank",
                style={"font-size": "20px"},
            ),
            "Link Generated",
            True,
        )

    @app.callback(
        Output("collapse-card", "is_open"),
        Input("collapse-button", "n_clicks"),
        State("collapse-card", "is_open"),
    )
    def toggle_collapse(n, is_open):
        if n:
            return not is_open
        return is_open

    @app.callback(
        Output("plot-collapse", "is_open"),
        Input("plot-collapse-button", "n_clicks"),
        State("plot-collapse", "is_open"),
    )
    def toggle_plot_collapse(n, is_open):
        if n:
            return not is_open
        return is_open

    pass


from typing import DefaultDict
from dash import html
from dash import dash_table
from dash import dcc
import dash_bootstrap_components as dbc
import flask

from ..common.dash_url_helper import create_component_kwargs, State

#####################
# title must be set #
#####################

# The title gives the title of the app in the browser tab
title = "Annotation Table Viewer"

###################################################################
# page_layout must be defined to take a state and return a layout #
###################################################################


def page_layout(state: State = {}):
    """This function returns the layout object for the dash app.
    The state parameter allows for URL-encoded parameter values.

    Parameters
    ----------
    state : State, optional
        URL state, a series of dicts that can provide parameter values, by default None

    Returns
    -------
    layout : list
        List of layout components for the dash app.
    """
    header_row = dbc.Row(
        [
            dbc.Col(
                html.Div(id='header-bar'),
                width={"size": 12},
            ),
        ],
    )

    cell_type_query = html.Div(
        children=[
            dbc.Row(
                [
                    dbc.Col(
                        [
                            html.Div("Annotation Table:"),
                            dcc.Dropdown(
                                **create_component_kwargs(
                                    state,
                                    id_inner="cell-type-table-menu",
                                    placeholder="Select a Table",
                                    options=[],
                                    value=None,
                                ),
                            ),
                        ],
                        width={"size": 3, "offset": 1},
                        align="end",
                    ),
                    dbc.Col(
                        [
                            dbc.Checklist(
                                **create_component_kwargs(
                                    state,
                                    id_inner="live-query-toggle",
                                    options=[
                                        {"label": "Live Query", "value": 1},
                                    ],
                                    value=[
                                        1,
                                    ],
                                    switch=True,
                                    style={
                                        "bottom-margin": "10px",
                                        "font-size": "16px",
                                    },
                                )
                            ),
                        ],
                        width={"size": 1, "offset": 0},
                        align="center",
                    ),
                    dbc.Col(
                        dbc.Button(
                            children="Submit",
                            id="submit-button",
                            color="primary",
                            style={"font-size": "18px"},
                        ),
                        width={"size": 1},
                        align="center",
                    ),
                    dbc.Col(
                        dcc.Loading(
                            id="main-loading",
                            children=html.Div(
                                id="main-loading-placeholder", children=""
                            ),
                            type="default",
                            style={"transform": "scale(0.8)"},
                        ),
                        align="end",
                    ),
                ],
                justify="start",
            ),
            dbc.Row(
                [
                    dbc.Col(
                        [
                            html.Div("Cell IDs (optional):"),
                            dbc.Input(
                                **create_component_kwargs(
                                    state,
                                    id_inner="anno-id",
                                    value="",
                                    type="text",
                                ),
                            ),
                        ],
                        width={"size": 2, "offset": 1},
                        align="end",
                    ),
                    dbc.Col(
                        [
                            dcc.Dropdown(
                                **create_component_kwargs(
                                    state,
                                    id_inner="id-type",
                                    options=[
                                        {
                                            "label": "Root ID",
                                            "value": "root_id",
                                        },
                                        {
                                            "label": "Nucleus ID",
                                            "value": "nucleus_id",
                                        },
                                        {
                                            "label": "Annotation ID",
                                            "value": "anno_id",
                                        },
                                    ],
                                    value="root_id",
                                    style={
                                        "margin-left": "12px",
                                        "font-size": "12px",
                                    },
                                    clearable=False,
                                )
                            ),
                        ],
                        width={"size": 1},
                        align="end",
                    ),
                    dbc.Col(
                        [html.Div(children="", id="report-text")],
                    ),
                ],
                justify="start",
            ),
            dbc.Row(
                [
                    dbc.Col(
                        [
                            
                            html.Div("Value Search (optional):"),
                            dbc.Input(
                                **create_component_kwargs(
                                    state,
                                    value="",
                                    id_inner="cell-type",
                                    type="text",
                                ),
                            ),
                        ],
                        width={"size": 2, "offset": 1},
                    ),
                    dbc.Col(
                        [
                            html.Div("Search Column:"),
                            dcc.Dropdown(
                                **create_component_kwargs(
                                    state,
                                    id_inner="value-column-search",
                                    options=[],
                                    value="",
                                    style={
                                        "margin-left": "12px",
                                        "font-size": "12px",
                                    },
                                    clearable=False,
                                )
                            ),
                        ],
                        width={"size": 1},
                        align="end",
                    )
                ],
                justify="state",
            ),
            html.Hr(),
        ]
    )

    message_row = dbc.Alert(
        id="message-text",
        children="Please select a cell type table and press Submit",
        color="info",
    )

    data_table = html.Div(
        dbc.Row(
            [
                dbc.Col(
                    dash_table.DataTable(
                        id="data-table",
                        columns=[{"id": "", "name": ""}],
                        data=[],
                        css=[
                            {
                                "selector": "table",
                                "rule": "table-layout: auto",
                            }
                        ],
                        style_cell={
                            "height": "auto",
                            "width": "12%",
                            "minWidth": "10%",
                            "maxWidth": "15%",
                            "whiteSpace": "normal",
                            "font-size": "11px",
                        },
                        style_header={"font-size": "12px", "fontWeight": "bold"},
                        sort_action="native",
                        sort_mode="multi",
                        filter_action="native",
                        row_selectable="multi",
                        page_current=0,
                        page_action="native",
                        page_size=50,
                        export_format="csv",
                        export_headers="names",
                    ),
                    width=10,
                )
            ],
            justify="center",
        )
    )

    ngl_link = dbc.Row(
        [
            dbc.Col(
                dbc.Spinner(
                    size="sm", children=html.Div(id="link-loading-placeholder")
                ),
                width=1,
                align="center",
            ),
            dbc.Col(
                dbc.Button(
                    children="Table View Neuroglancer Link",
                    id="ngl-link",
                    href="",
                    target="_blank",
                    style={"font-size": "16px"},
                    color="primary",
                    external_link=True,
                    disabled=False,
                ),
                width=3,
                align="start",
            ),
            dbc.Col(
                dbc.Button(
                    id="reset-selection",
                    children="Reset Selection",
                    color="warning",
                    size="sm",
                ),
                width=1,
            ),
            dbc.Col(
                html.A(
                    "Instructions for filtering the table",
                    href="https://dash.plotly.com/datatable/filtering",
                    style={"font-size": "15px"},
                    target="_blank",
                ),
                align="center",
                width={"size": 2, "offset": 3},
            ),
        ],
        justify="start",
    )

    whole_column_link = dbc.Row(
        [
            dbc.Col(
                dbc.Card(
                    [
                        dbc.CardBody(
                            [
                                html.H4("Whole Table Link", className="card-title"),
                                html.Div(
                                    children=[
                                        dbc.Button(
                                            "Generate Link",
                                            id="whole-table-link-button",
                                            color="secondary",
                                            className="d-grid gap-2 col-6 mx-auto",
                                        ),
                                    ]
                                ),
                                dbc.Spinner(
                                    dbc.Row(
                                        html.Div(
                                            "", id="whole-table-link", className="card-text"
                                        ),
                                        justify='center',
                                        align='center',
                                    ),
                                    size="sm",
                                ),
                            ]
                        )
                    ]
                ),
            ),
            dbc.Col(
                dbc.Card(
                    dbc.CardBody(
                        [
                            html.H4('Annotation Grouping', className='card-title'),
                            dbc.Checklist(
                                options=[
                                    {"label": "Group annotations", "value": 1},
                                ],
                                value=[],
                                id="do-group",
                                switch=True,
                            ),
                            dcc.Dropdown(
                                options={},
                                value='cell_type',
                                id='group-by',
                                searchable=False,
                            ),
                        ]
                    )
                )
            )
        ],
    )

    datastack_comp = (
        dcc.Input(
            **create_component_kwargs(
                state,
                id_inner="datastack",
                value="",
                type="text",
            ),
        ),
    )

    title_row = dbc.Row(
        [
            dbc.Col(
                html.H3("Table Viewer"),
                width={"size": 6, "offset": 1},
            ),
        ],
    )

    layout = html.Div(
        children=[
            header_row,
            dbc.Container(cell_type_query, fluid=True),
            dbc.Container(message_row),
            dbc.Container(whole_column_link),
            html.Hr(),
            html.Div(ngl_link),
            html.Div(data_table),
            html.Div(datastack_comp, style={"display": "none"}),
            dcc.Store(id="client-info-json"),
            dcc.Store(id="table-resolution-json"),
            dcc.Store(id='data-resolution-json'),
            dcc.Store(id='pt-column'),
            dcc.Store(id='value-columns'),
        ]
    )
    return layout


######################################################
# Leave this rest alone for making the template work #
######################################################

url_bar_and_content_div = html.Div(
    [dcc.Location(id="url", refresh=False), html.Div(id="page-layout")]
)


def app_layout():
    # https://dash.plotly.com/urls "Dynamically Create a Layout for Multi-Page App Validation"
    if flask.has_request_context():  # for real
        return url_bar_and_content_div
    # validation only
    return html.Div([url_bar_and_content_div, *page_layout()])



# from https://gist.github.com/fzyzcjy/0322eebd54d4889b03e0c3ea9fd9e965
import ast
import re
from typing import Dict, Callable, Any
from urllib.parse import urlparse, parse_qsl, urlencode, quote

import dash
from dash.dependencies import Input, Output, ALL

_COMPONENT_ID_TYPE = 'url_helper'

"""
definition: {id_inner: {property: your_value}}
NOTE here we use id_inner, NOT the real id (which is a dict)
"""
State = Dict[str, Dict[str, Any]]


def create_component_kwargs(state: State, id_inner: str, **raw_kwargs) -> Dict[str, Any]:
    # noinspection PyDictCreation
    kwargs = {**raw_kwargs}

    # create "id"
    kwargs['id'] = {
        'type': _COMPONENT_ID_TYPE,
        'id_inner': id_inner,
    }

    # apply default value
    if id_inner in state:
        param_key_dict = state[id_inner]
        kwargs.update(param_key_dict)

    return kwargs


_ID_PARAM_SEP = '::'


def _parse_url_to_state(href: str) -> State:
    parse_result = urlparse(href)
    query_string = parse_qsl(parse_result.query)

    state = {}
    for key, value in query_string:
        if _ID_PARAM_SEP in key:
            id, param = key.split(_ID_PARAM_SEP)
        else:
            id, param = key, 'value'
        try:
            state.setdefault(id, {})[param] = ast.literal_eval(value)
        except SyntaxError:
            state.setdefault(id, {})[param] = value

    return state


def _param_string(id_inner: str, property: str) -> str:
    return id_inner if property == 'value' else id_inner + _ID_PARAM_SEP + property


_RE_SINGLE_QUOTED = re.compile("^'|'$")


def _myrepr(o: str) -> str:
    """Optional but chrome URL bar hates "'" """
    # p.s. Pattern.sub(repl, string)
    return _RE_SINGLE_QUOTED.sub('"', repr(o))


def setup(app: dash.Dash, page_layout: Callable[[State], Any]):
    """
    NOTE ref: https://github.com/plotly/dash/issues/188
    """

    @app.callback(
        Output('page-layout', 'children'),
        inputs=[Input('url', 'href')],
    )
    def page_load(href: str):
        if not href:
            return []

        state = _parse_url_to_state(href)
        return page_layout(state)

    @app.callback(
        Output('url', 'search'),
        # NOTE currently only support property="value"...
        Input({'type': _COMPONENT_ID_TYPE, 'id_inner': ALL}, 'value'),
    )
    def update_url_state(values):
        """Updates URL from component values."""

        state = {}

        # https://dash.plotly.com/pattern-matching-callbacks
        inputs = dash.callback_context.inputs_list[0]
        for input in inputs:
            id = input['id']
            assert isinstance(id, Dict)
            assert id['type'] == _COMPONENT_ID_TYPE
            id_inner = id['id_inner']
            state[_param_string(id_inner, input['property'])] = _myrepr(input['value'])

        params = urlencode(state, safe="%/:?~#+!$,;'@()*[]\"", quote_via=quote)

        return f'?{params}'


from concurrent.futures import ThreadPoolExecutor
from .schema_utils import table_metadata
import pandas as pd
import re
import numpy as np
from .transform_utils import extract_depth

DESIRED_RESOLUTION = [1, 1, 1]


def query_table_any(
    table, root_id_column, root_ids, client, timestamp, extra_query={}, is_live=True
):
    if root_ids is not None:
        root_ids = np.array(root_ids)
        root_ids = root_ids[root_ids != 0]
    ref_table = table_metadata(table, client).get("reference_table")
    if ref_table is not None:
        return _query_table_join(
            table,
            root_id_column,
            root_ids,
            client,
            timestamp,
            ref_table,
            extra_query=extra_query,
            is_live=is_live,
        )
    else:
        return _query_table_single(
            table,
            root_id_column,
            root_ids,
            client,
            timestamp,
            extra_query=extra_query,
            is_live=is_live,
        )


def _query_table_single(
    table, root_id_column, root_ids, client, timestamp, extra_query, is_live
):
    filter_kwargs = {}
    if root_ids is not None:
        if len(root_ids) == 1:
            if is_live:
                filter_kwargs["filter_equal_dict"] = {
                    table: {root_id_column: root_ids[0]}
                }
            else:
                filter_kwargs["filter_equal_dict"] = {root_id_column: root_ids[0]}
        else:
            if is_live:
                filter_kwargs["filter_in_dict"] = {table: {root_id_column: root_ids}}
            else:
                filter_kwargs["filter_in_dict"] = {root_id_column: root_ids}
    if len(extra_query) != 0:
        if "filter_in_dict" in filter_kwargs:
            if is_live:
                filter_kwargs["filter_in_dict"][table].extend(extra_query)
            else:
                filter_kwargs["filter_in_dict"].extend(extra_query)
        else:
            if is_live:
                filter_kwargs["filter_in_dict"] = {table: extra_query}
            else:
                filter_kwargs["filter_in_dict"] = extra_query
    if is_live:
        return client.materialize.live_live_query(
            table,
            timestamp=timestamp,
            split_positions=True,
            desired_resolution=DESIRED_RESOLUTION,
            allow_missing_lookups=True,
            metadata=False,
            **filter_kwargs,
        )
    else:
        return client.materialize.query_table(
            table,
            split_positions=True,
            desired_resolution=DESIRED_RESOLUTION,
            metadata=False,
            **filter_kwargs,
        )


def _query_table_join(
    table, root_id_column, root_ids, client, timestamp, ref_table, extra_query, is_live
):
    filter_kwargs = {}
    if root_ids is not None:
        if len(root_ids) == 1:
            filter_kwargs = {
                "filter_equal_dict": {ref_table: {root_id_column: root_ids[0]}}
            }
        else:
            filter_kwargs = {"filter_in_dict": {ref_table: {root_id_column: root_ids}}}
    if len(extra_query) != 0:
        if "filter_in_dict" in filter_kwargs:
            filter_kwargs["filter_in_dict"][table].extend(extra_query)
        else:
            filter_kwargs["filter_in_dict"] = {table: extra_query}
    if is_live:
        join = [[table, "target_id", ref_table, "id"]]
        return client.materialize.live_live_query(
            table,
            joins=join,
            timestamp=timestamp,
            split_positions=True,
            desired_resolution=DESIRED_RESOLUTION,
            suffixes={table: "", ref_table: "_ref"},
            allow_missing_lookups=True,
            metadata=False,
            **filter_kwargs,
        ).rename(columns={"idx": "id"})
    else:
        join = [[table, "target_id"], [ref_table, "id"]]
        return client.materialize.join_query(
            join,
            suffixes={table: "", ref_table: "_ref"},
            split_positions=True,
            desired_resolution=DESIRED_RESOLUTION,
            metadata=False,
            **filter_kwargs,
        )


def get_specific_soma(soma_table, root_id, client, timestamp, is_live):
    soma_df = query_table_any(
        soma_table, "pt_root_id", [root_id], client, timestamp, is_live=is_live
    )
    return soma_df


def _synapse_df(
    direction,
    synapse_table,
    root_id,
    client,
    timestamp,
    synapse_position_column,
    synapse_table_columns,
    exclude_autapses=True,
    is_live=True,
):
    if is_live:
        syn_df = client.materialize.query_table(
            synapse_table,
            filter_equal_dict={f"{direction}_pt_root_id": root_id},
            split_positions=True,
            timestamp=timestamp,
            desired_resolution=DESIRED_RESOLUTION,
            metadata=False,
        )
    else:
        syn_df = client.materialize.query_table(
            synapse_table,
            filter_equal_dict={f"{direction}_pt_root_id": root_id},
            split_positions=True,
            desired_resolution=DESIRED_RESOLUTION,
            metadata=False,
        )

    if exclude_autapses:
        syn_df = syn_df.query("pre_pt_root_id != post_pt_root_id").reset_index(
            drop=True
        )
    return syn_df[synapse_table_columns]


def pre_synapse_df(
    synapse_table,
    root_id,
    client,
    timestamp,
    config,
    is_live,
):
    return _synapse_df(
        "pre",
        synapse_table,
        root_id,
        client,
        timestamp,
        config.syn_pt_position,
        config.synapse_table_columns_dataframe,
        is_live=is_live,
    )


def post_synapse_df(synapse_table, root_id, client, timestamp, config, is_live):
    return _synapse_df(
        "post",
        synapse_table,
        root_id,
        client,
        timestamp,
        config.syn_pt_position,
        config.synapse_table_columns_dataframe,
        is_live=is_live,
    )


def synapse_data(
    synapse_table,
    root_id,
    client,
    timestamp,
    config,
    n_threads=2,
    is_live=True,
):
    if n_threads > 2:
        n_threads = 2
    with ThreadPoolExecutor(n_threads) as exe:
        pre = exe.submit(
            pre_synapse_df,
            synapse_table,
            root_id,
            client,
            timestamp,
            config,
            is_live,
        )
        post = exe.submit(
            post_synapse_df,
            synapse_table,
            root_id,
            client,
            timestamp,
            config,
            is_live,
        )
    return pre.result(), post.result()


def stringify_root_ids(df, stringify_cols=None):
    if stringify_cols is None:
        stringify_cols = [col for col in df.columns if re.search("_root_id$", col)]
    for col in stringify_cols:
        df[col] = df[col].astype(str)
    return df


def stringify_list(col, df):
    df[col] = df[col].apply(lambda x: str(x)[1:-1]).astype(str)
    return df


def repopulate_list(col, df):
    df[col] = df[col].apply(lambda x: [float(y) for y in x.split(",")]).astype(object)


def rehydrate_dataframe(rows, columns=[]):
    df = pd.DataFrame(rows)
    for col in columns:
        repopulate_list(col, df)
    return df


def _expand_column(df, column, len_column="num_syn"):
    def _expand_column(row):
        return [row[column]] * row[len_column]

    return np.concatenate(list(df.apply(_expand_column, axis=1)))


def _slam_column(df, column):
    return np.concatenate(df[column].values)


def rebuild_synapse_dataframe(rows, config, aligned_volume, value_cols=[]):
    if len(rows) == 0:
        return pd.DataFrame(columns=config.syn_pt_position_split + value_cols)

    df = rehydrate_dataframe(rows, config.syn_pt_position_split).replace("nan", None)
    dfnn = df.dropna(subset=config.soma_depth_column)
    data_dict = {k: _slam_column(dfnn, k) for k in config.syn_pt_position_split}

    value_cols.append(config.soma_depth_column)
    for col in value_cols:
        if col != "":
            data_dict[col] = _expand_column(dfnn, col)
    df_rh = pd.DataFrame(data_dict).replace("nan", None)

    extract_depth(
        df_rh,
        config.synapse_depth_column,
        config.syn_pt_position,
        aligned_volume,
    )
    return df_rh


def _get_single_table(
    table_name,
    root_ids,
    root_id_column,
    include_columns,
    aggregate_map,
    client,
    timestamp,
    table_filter=None,
    is_live=True,
):
    keep_columns = include_columns.copy()
    df = query_table_any(
        table_name, root_id_column, root_ids, client, timestamp, is_live=is_live
    )
    if table_filter is not None:
        df = df.query(table_filter).reset_index(drop=True)

    for k, v in aggregate_map.items():
        df[k] = df.groupby(v["group_by"])[v["column"]].transform(v["agg"])
        keep_columns.append(k)
    if len(aggregate_map) != 0:
        is_dup = df.duplicated(root_id_column, False)
        if np.any(is_dup):
            df.loc[df.index[is_dup], include_columns] = np.nan
        df.drop_duplicates(root_id_column, keep="first", inplace=True)
    else:
        df.drop_duplicates(root_id_column, keep=False, inplace=True)
    df.set_index(root_id_column, inplace=True)
    return df[keep_columns]


def property_table_data(
    root_ids,
    property_mapping,
    client,
    timestamp,
    n_threads=2,
    is_live=True,
):
    if len(property_mapping) == 0:
        return {}

    jobs = []
    with ThreadPoolExecutor(n_threads) as exe:
        for table_name, attrs in property_mapping.items():
            jobs.append(
                exe.submit(
                    _get_single_table,
                    table_name,
                    root_ids,
                    attrs.get("root_id"),
                    attrs.get("include", []),
                    attrs.get("aggregate", {}),
                    client,
                    timestamp,
                    attrs.get("table_filter", None),
                    is_live,
                )
            )
    return {tname: job.result() for tname, job in zip(property_mapping, jobs)}



from dash import dash_table
from dash import dcc
from dash import html
import dash_bootstrap_components as dbc
import flask
from ..common.dash_url_helper import create_component_kwargs, State

title = "Synapse Table Viewer"

url_bar_and_content_div = html.Div(
    [dcc.Location(id="url", refresh=False), html.Div(id="page-layout")]
)


def page_layout(state: State = None):
    state = state or {}
    
    header_row = dbc.Row(
        [
            dbc.Col(
                html.Div(id='header-bar'),
                width={"size": 12},
            ),
        ],
    )


    input_row = [
        dbc.Row(
            [
                dbc.Col(
                    [
                        html.Div("Cell ID:"),
                    ],
                    align="end",
                ),
            ],
            justify="start",
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        dbc.Input(
                            **create_component_kwargs(
                                state,
                                id_inner="anno-id",
                                value="",
                                type="text",
                                # style={"font-size": "18px"},
                            )
                        ),
                    ],
                    align="start",
                ),
                dbc.Col(
                    [
                        dbc.Checklist(
                            **create_component_kwargs(
                                state,
                                id_inner="live-query-toggle",
                                options=[
                                    {"label": "Live Query", "value": 1},
                                ],
                                value=[
                                    1,
                                ],
                                switch=True,
                            )
                        ),
                    ],
                    align="center",
                ),
                dbc.Col(
                    [
                        dbc.Button(
                            id="submit-button",
                            children="Submit",
                            color="primary",
                            style={"font-size": "16px", "align": "left"},
                        ),
                    ],
                    align="start",
                ),
                dbc.Col(
                    [
                        html.Div(
                            dcc.Loading(
                                id="main-loading",
                                children=html.Div(id="loading-spinner", children=""),
                                style={"transform": "scale(1)"},
                                type="default",
                            )
                        )
                    ],
                    align="center",
                ),
            ],
            justify="start",
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        dcc.Dropdown(
                            **create_component_kwargs(
                                state,
                                id_inner="cell-id-type",
                                options=[
                                    {"label": "Root ID", "value": "root_id"},
                                    {"label": "Nucleus ID", "value": "nucleus_id"},
                                ],
                                value="root_id",
                                clearable=False,
                            )
                        ),
                    ],
                    width=2,
                    align="end",
                ),
            ],
            justify="start",
        ),
    ]

    message_row = dbc.Alert(
        id="message-text",
        children="Please select a neuron id",
        color="info",
    )

    data_table = html.Div(
        [
            dcc.Tabs(
                id="connectivity-tab",
                value="tab-pre",
                children=[
                    dcc.Tab(id="input-tab", label="Input", value="tab-post"),
                    dcc.Tab(id="output-tab", label="Output", value="tab-pre"),
                ],
            ),
            html.Div(
                dbc.Row(
                    [
                        dbc.Col(
                            dash_table.DataTable(
                                id="data-table",
                                columns=[{"id": "", "name": ""}],
                                data=[],
                                css=[
                                    {
                                        "selector": "table",
                                        "rule": "table-layout: fixed",
                                    }
                                ],
                                style_cell={
                                    "height": "auto",
                                    "whiteSpace": "normal",
                                    "font-size": "11px",
                                },
                                style_header={
                                    "font-size": "12px",
                                    "fontWeight": "bold",
                                },
                                sort_action="native",
                                sort_mode="multi",
                                filter_action="native",
                                row_selectable="multi",
                                page_current=0,
                                page_action="native",
                                page_size=50,
                                export_format="csv",
                                export_headers="names",
                            ),
                            width=10,
                        ),
                    ],
                    justify="center",
                )
            ),
        ]
    )

    cell_links = html.Div(
        [
            dbc.Row(
                [
                    dbc.Col(
                        dbc.Button(
                            "Generate All Input Link",
                            id="all-input-link-button",
                            color="secondary",
                        ),
                    ),
                    dbc.Col(
                        dbc.Button(
                            "Generate All Output Link",
                            id="all-output-link-button",
                            color="secondary",
                        )
                    ),
                ]
            ),
            dbc.Row(
                [
                    dbc.Col(
                        html.Div("", id="all-input-link"),
                    ),
                    dbc.Col(html.Div("", id="all-output-link")),
                ]
            ),
        ]
    )

    top_link = dbc.Row(
        [
            dbc.Col(
                dbc.Spinner(size="sm", children=html.Div(id="link-loading")),
                width=1,
                align="center",
            ),
            dbc.Col(
                [
                    dbc.Button(
                        children="Neuroglancer Link",
                        color="primary",
                        external_link=True,
                        target="_blank",
                        id="ngl_link",
                        href="",
                        disabled=False,
                    ),
                ],
            ),
            dbc.Col(
                dbc.Button(
                    id="reset-selection",
                    children="Reset Selection",
                    color="warning",
                    size="sm",
                ),
            ),
        ],
        justify="left",
    )

    layout = html.Div(
        children=[
            header_row,
            dbc.Container(
                input_row,
                fluid=True,
            ),
            html.Hr(),
            dbc.Container(message_row),
            # html.Hr(),
            dbc.Container(cell_links),
            html.Hr(),
            dbc.Container(top_link, fluid=True),
            data_table,
            dcc.Store("target-table-json"),
            dcc.Store("source-table-json"),
            dcc.Store("client-info-json"),
            dcc.Store("synapse-table-resolution-json"),
            html.Div(
                dcc.Input(
                    **create_component_kwargs(
                        state,
                        id_inner="datastack",
                        value="",
                    ),
                ),
                style={"display": "none"},
            ),
        ]
    )

    return layout


def app_layout():
    # https://dash.plotly.com/urls "Dynamically Create a Layout for Multi-Page App Validation"
    if flask.has_request_context():  # for real
        return url_bar_and_content_div
    # validation only
    return html.Div([url_bar_and_content_div, *page_layout()])



from ..common.config import CommonConfig, RegisterTable
from ..common.schema_utils import bound_pt_position, bound_pt_root_id, split_pt_position
import copy


class CellTypeConfig(CommonConfig):
    def __init__(self, config):
        super().__init__(config)

        self.allowed_cell_type_schema_bridge = config.get("ct_cell_type_schema", {})
        self.ct_cell_type_point = None
        self.value_columns = []
        self.omit_cell_type_tables = config.get("omit_cell_type_tables", [])
        self.cell_type_dropdown_options = config.get("cell_type_dropdown_options", [])

    @property
    def ct_cell_type_pt_position(self):
        if self.ct_cell_type_point is None:
            return None
        return bound_pt_position(self.ct_cell_type_point)

    @property
    def ct_cell_type_pt_position_split(self):
        if self.ct_cell_type_point is []:
            return None
        return split_pt_position(self.ct_cell_type_pt_position)

    @property
    def ct_cell_type_root_id(self):
        if self.ct_cell_type_point is None:
            return None
        return bound_pt_root_id(self.ct_cell_type_point)

    @property
    def ct_table_columns(self):
        return (
            [
                self.root_id_col,
            ]
            + self.value_columns
            + self.ct_cell_type_pt_position_split
        )


import flask
import datetime
from dash import callback_context
from dash import dcc
from dash import html
from .config import CellTypeConfig, RegisterTable

from dash.dependencies import Input, Output, State
from ..common.dataframe_utilities import *
from ..common.link_utilities import (
    DEFAULT_NGL,
    generate_statebuilder,
    generate_url_cell_types,
    EMPTY_INFO_CACHE,
    MAX_URL_LENGTH,
)
from ..common.lookup_utilities import (
    get_type_tables,
    make_client,
)
from ..common.schema_utils import get_table_info
from ..common.table_lookup import TableViewer

# Callbacks using data from URL-encoded parameters requires this import
from ..common.dash_url_helper import _COMPONENT_ID_TYPE

InputDatastack = Input({"id_inner": "datastack", "type": _COMPONENT_ID_TYPE}, "value")
InputCellTypeMenu = Input(
    {"id_inner": "cell-type-table-menu", "type": _COMPONENT_ID_TYPE}, "value"
)

OutputDatastack = Output({"id_inner": "datastack", "type": _COMPONENT_ID_TYPE}, "value")
OutputCellTypeMenuOptions = Output(
    {"id_inner": "cell-type-table-menu", "type": _COMPONENT_ID_TYPE}, "options"
)
StateCellTypeMenu = State(
    {"id_inner": "cell-type-table-menu", "type": _COMPONENT_ID_TYPE}, "value"
)
StateCellType = State({"id_inner": "cell-type", "type": _COMPONENT_ID_TYPE}, "value")
StateAnnoID = State({"id_inner": "anno-id", "type": _COMPONENT_ID_TYPE}, "value")
StateCategoryID = State({"id_inner": "id-type", "type": _COMPONENT_ID_TYPE}, "value")
StateLiveQuery = State(
    {"id_inner": "live-query-toggle", "type": _COMPONENT_ID_TYPE}, "value"
)
StateValueSearch = State(
    {"id_inner": "value-column-search", "type": _COMPONENT_ID_TYPE}, "value"
)

OutputLiveQueryToggle = Output(
    {"id_inner": "live-query-toggle", "type": _COMPONENT_ID_TYPE},
    "options",
)
OutputLiveQueryValue = Output(
    {"id_inner": "live-query-toggle", "type": _COMPONENT_ID_TYPE}, "value"
)
OutputValueSearch = Output(
    {"id_inner": "value-column-search", "type": _COMPONENT_ID_TYPE}, "options"
)

######################################
# register_callbacks must be defined #
######################################


def register_callbacks(app, config):
    """This function must be present and add all callbacks to the app.
    Note that inputs from url-encoded values have a different structure than other values.
    A config dict is also allowed to configure standard parameter values for use in callback functions.

    Here, we show basic examples of using the three parameters defined in the layout.page_layout function.

    Parameters
    ----------
    app : Dash app
        Pre-made dash app
    config : dict
        Dict for standard parameter values
    """
    c = CellTypeConfig(config)

    @app.callback(
        OutputDatastack,
        InputDatastack,
    )
    def define_datastack(datastack):
        if datastack is None:
            datastack = ""

        if len(datastack) == 0:
            return c.default_datastack
        else:
            return datastack

    @app.callback(
        Output("header-bar", 'children'),
        InputDatastack,
    )
    def set_header(datastack):
        return html.H3(f"Table Info — {datastack}", className="bg-primary text-white p-2 mb-2 text-center")

    @app.callback(
        OutputCellTypeMenuOptions,
        InputDatastack,
    )
    def cell_type_dropdown(datastack):
        return get_type_tables(datastack, c)

    @app.callback(
        OutputLiveQueryToggle,
        OutputLiveQueryValue,
        InputDatastack,
        StateLiveQuery,
    )
    def disable_live_query(_, lq):
        options_active = [{"label": "Live Query", "value": 1}]
        options_disabled = [{"label": "Live Query", "value": 1, "disabled": True}]
        if c.disallow_live_query:
            return options_disabled, ""
        else:
            return options_active, lq

    @app.callback(
        OutputValueSearch,
        InputDatastack,
        InputCellTypeMenu,
    )
    def update_value_search_list(datastack, table_name):
        if table_name is None:
            return []
        client = make_client(datastack, c.server_address)
        _, cols = get_table_info(table_name, client, merge_schema=False)
        return [{"label": i, "value": i} for i in cols]


    @app.callback(
        Output("group-by", "options"),
        Input("submit-button", "n_clicks"),
        StateCellTypeMenu,
        InputDatastack,
    )
    def update_groupby_list(_, cell_type_table, datastack):
        if cell_type_table == "" or cell_type_table is None:
            return {}
        else:
            client = make_client(datastack, c.server_address)
            _, cols = get_table_info(cell_type_table, client, allow_types=['boolean', 'integer', 'string'])
            return {k: k for k in cols}

    @app.callback(
        Output("data-table", "columns"),
        Output("pt-column", "data"),
        Output("value-columns", "data"),
        Input("submit-button", "n_clicks"),
        InputDatastack,
        StateCellTypeMenu,
    )
    def define_table_columns(_, datastack, cell_type_table):
        client = make_client(datastack, c.server_address)
        pt, cols = get_table_info(cell_type_table, client)
        reg_con = RegisterTable(pt, cols, c)
        return (
            [{"name": i, "id": i} for i in reg_con.ct_table_columns],
            pt,
            cols,
        )

    @app.callback(
        Output("data-table", "data"),
        Output("message-text", "children"),
        Output("main-loading-placeholder", "value"),
        Output("client-info-json", "data"),
        Output("message-text", "color"),
        Output("data-resolution-json", "data"),
        Input("submit-button", "n_clicks"),
        InputDatastack,
        StateCellTypeMenu,
        StateAnnoID,
        StateCategoryID,
        StateCellType,
        StateLiveQuery,
        StateValueSearch,
    )
    def update_table(
        clicks,
        datastack,
        cell_type_table,
        anno_id,
        id_type,
        value_search,
        live_query_toggle,
        value_search_field,
    ):
        try:
            client = make_client(datastack, c.server_address)
            info_cache = client.info.get_datastack_info()
            info_cache["global_server"] = client.server_address
        except Exception as e:
            return [], str(e), "", EMPTY_INFO_CACHE, "danger", c.data_resolution

        if cell_type_table is None:
            return [], "No Table Selected", "", info_cache, "info", c.data_resolution

        if len(anno_id) == 0:
            anno_id = None
        else:
            anno_id = [int(x) for x in anno_id.split(",")]

        live_query = len(live_query_toggle) == 1

        if live_query and not c.disallow_live_query:
            timestamp = datetime.datetime.utcnow()
        else:
            timestamp = None
            timestamp_ngl = client.materialize.get_timestamp()
            info_cache["ngl_timestamp"] = timestamp_ngl.timestamp()

        anno_type_lookup = {
            "root_id": "root",
            "nucleus_id": "nucleus",
            "anno_id": "annotation",
        }

        annotation_filter = {}
        if value_search is not None or value_search_field is not None:
            if len(value_search_field) > 0 and len(value_search) > 0:
                annotation_filter = {value_search_field: value_search.split(',')}

        try:
            tv = TableViewer(
                cell_type_table,
                client,
                c,
                id_query=anno_id,
                id_query_type=anno_type_lookup[id_type],
                column_query=annotation_filter,
                timestamp=timestamp,
                is_live=live_query,
            )
            df = tv.table_data()
            if live_query:
                output_report = f"Current state of table {cell_type_table}"
            else:
                output_report = f"Table {cell_type_table} materialized on {timestamp_ngl:%m/%d/%Y} (v{client.materialize.version})"
            output_color = "success"
        except Exception as e:
            df = pd.DataFrame(columns=c.ct_table_columns)
            output_report = str(e)
            output_color = "danger"

        ct_df = stringify_root_ids(df, stringify_cols=[c.root_id_col])

        return (
            ct_df.to_dict("records"),
            output_report,
            "",
            info_cache,
            output_color,
            c.data_resolution,
        )

    @app.callback(
        Output("data-table", "selected_rows"),
        Input("reset-selection", "n_clicks"),
    )
    def reset_selection(n_clicks):
        return []

    @app.callback(
        Output("ngl-link", "href"),
        Output("ngl-link", "children"),
        Output("ngl-link", "disabled"),
        Output("link-loading-placeholder", "children"),
        Input("data-table", "derived_virtual_data"),
        Input("data-table", "derived_virtual_selected_rows"),
        Input("client-info-json", "data"),
        Input("data-resolution-json", "data"),
        Input("pt-column", "data"),
        Input("do-group", "value"),
        Input("group-by", "value"),
    )
    def update_link(
        rows,
        selected_rows,
        info_cache,
        data_resolution,
        pt_column,
        do_group,
        group_column,
    ):
        def state_text(n):
            return f"Neuroglancer: ({n} rows)"

        if info_cache is None:
            return "", "No datastack set", True, ""

        if pt_column is None:
            return "", "No clear point field in table", True, ""

        if 1 in do_group:
            do_group = True
        else:
            do_group = False

        if group_column is None:
            do_group = False
        else:
            if len(group_column) == 0:
                do_group = False

        if rows is None or len(rows) == 0:
            sb = generate_statebuilder(info_cache, c, anno_layer="anno")
            url = sb.render_state(None, return_as="url")
            link_name = state_text(0)
            link_color = True
        else:
            df = pd.DataFrame(rows)
            if len(df) > c.max_dataframe_length and len(selected_rows) == 0:
                url = ""
                link_name = "State Too Large"
                link_color = True
            else:
                url = generate_url_cell_types(
                    selected_rows,
                    df,
                    info_cache,
                    c,
                    pt_column,
                    group_annotations=do_group,
                    cell_type_column=group_column,
                    data_resolution=data_resolution,
                )
                if len(url) > MAX_URL_LENGTH:
                    url = ""
                    link_name = "State Too Large"
                    link_color = True
                else:
                    if len(selected_rows) == 0:
                        link_name = state_text(len(df))
                    else:
                        link_name = state_text(len(selected_rows))
                    link_color = False
        return url, link_name, link_color, ""

    @app.callback(
        Output("whole-table-link", "children"),
        Output("whole-table-link-button", "children"),
        Output("whole-table-link-button", "disabled"),
        Input("whole-table-link-button", "n_clicks"),
        Input("submit-button", "n_clicks"),
        Input("data-table", "data"),
        Input("client-info-json", "data"),
        InputDatastack,
        Input("data-resolution-json", "data"),
        Input("pt-column", "data"),
        Input("do-group", "value"),
        Input("group-by", "value"),
        prevent_initial_call=True,
    )
    def update_whole_table_link(_1, _2, rows, info_cache, datastack, data_resolution, pt_column, do_group, group_column):
        ctx = callback_context
        if not ctx.triggered:
            return ""
        trigger_src = ctx.triggered[0]["prop_id"].split(".")[0]
        if trigger_src in [
            "submit-button",
            "client-info-json",
            "data-table",
            "pt-column",
            "do-group",
            "group-by",
        ]:
            return "", "Generate Link", False

        if rows is None or len(rows) == 0:
            return html.Div("No items to show"), "Error", True

        if pt_column is None:
            return "", "No clear point field in table", True, ""

        if 1 in do_group:
            do_group = True
        else:
            do_group = False

        if group_column is None:
            do_group = False
        else:
            if len(group_column) == 0:
                do_group = False

        df = pd.DataFrame(rows)
        if len(df) > c.max_server_dataframe_length:
            df = df.sample(c.max_server_dataframe_length)
            sampled = True
        else:
            sampled = False

        if len(df) > c.max_dataframe_length:
            try:
                client = make_client(datastack, c.server_address)
                state = generate_url_cell_types(
                    [],
                    df,
                    info_cache,
                    c,
                    pt_column,
                    group_annotations=do_group,
                    cell_type_column=group_column,
                    data_resolution=data_resolution,
                    return_as="dict",
                )
                state_id = client.state.upload_state_json(state)
                ngl_url = client.info.viewer_site()
                if ngl_url is None:
                    ngl_url = DEFAULT_NGL
                url = client.state.build_neuroglancer_url(state_id, ngl_url=ngl_url)
            except Exception as e:
                return html.Div(str(e)), "Error", True
        else:
            url = generate_url_cell_types(
                [],
                df,
                info_cache,
                c,
                pt_column,
                group_annotations=do_group,
                cell_type_column=group_column,
                data_resolution= data_resolution,
            )

        if sampled:
            link_text = f"Neuroglancer Link (State very large — Random {c.max_server_dataframe_length} shown)"
        else:
            link_text = f"Neuroglancer Link"

        return (
            html.A(link_text, href=url, target="_blank", style={"font-size": "20px"}),
            "Link Generated",
            True,
        )

    pass



import flask
from .schema_utils import get_table_info, populate_metadata_cache
from caveclient.tools.caching import CachedClient as CAVEclient
from .dataframe_utilities import query_table_any
import numpy as np
# from concurrent.futures import ThreadPoolExecutor

def table_is_value_source(table, client):
    if table is None:
        return False
    pt, vals = get_table_info(table, client)
    if pt is not None and len(vals) > 0:
        return True
    else:
        return False

def get_all_schema_tables(
    datastack,
    config,
):
    client = make_client(datastack, config.server_address)
    tables = client.materialize.get_tables()
    populate_metadata_cache(tables, client)
    schema_tables = []
    is_val_source = {t: table_is_value_source(t, client) for t in tables if t not in config.omit_cell_type_tables}
    schema_tables = [k for k, v in is_val_source.items() if v]
    return [{"label": t, "value": t} for t in sorted(schema_tables)]

def get_type_tables(datastack, config):
    tables = get_all_schema_tables(datastack, config)

    named_options = config.cell_type_dropdown_options
    if named_options is None:
        return tables
    else:
        if len(named_options) == 0:
            named_option_dict = dict()
        named_option_dict = {r["value"]: r["label"] for r in named_options[::-1]}

    new_tables = []
    for t in tables:
        if t["value"] in named_option_dict:
            new_tables = [
                {"label": named_option_dict.get(t["value"]), "value": t["value"]}
            ] + new_tables
        else:
            new_tables.append(t)
    return new_tables


def make_client(datastack, server_address, **kwargs):
    """Build a framework client with appropriate auth token

    Parameters
    ----------
    datastack : str
        Datastack name for client
    config : dict
        Config dict for settings such as server address.
    server_address : str, optional
        Global server address for the client, by default None. If None, uses the config dict.

    """
    try:
        auth_token = flask.g.get("auth_token", None)
    except:
        auth_token = None
    client = CAVEclient(datastack, server_address=server_address, auth_token=auth_token, **kwargs)
    return client


def get_root_id_from_nuc_id(
    nuc_id,
    client,
    nucleus_table,
    config,
    timestamp=None,
    is_live=True,
):
    """Look up current root id from a nucleus id

    Parameters
    ----------
    nuc_id : int
        Annotation id from a nucleus
    client : CAVEclient
        CAVEclient for the server in question
    nucleus_table : str
        Name of the table whose annotation ids are nucleus lookups.
    timestamp : datetime.datetime, optional
        Timestamp for live query lookup. Required if live is True. Default is None.
    live : bool, optional
        If True, uses a live query. If False, uses the materialization version set in the client.

    Returns
    -------
    [type]
        [description]
    """
    df = query_table_any(
            nucleus_table,
            config.soma_pt_root_id,
            None,
            client,
            timestamp=timestamp,
            extra_query={config.nucleus_id_column: [nuc_id]},
            is_live=is_live,
    )
    if len(df) == 0:
        return None
    else:
        return df.iloc[0][config.soma_pt_root_id]


def get_nucleus_id_from_root_id(
    root_id,
    client,
    nucleus_table,
    config,
    timestamp=None,
    is_live=True,
):
    df = query_table_any(
        nucleus_table,
        config.soma_pt_root_id,
        np.array([root_id]),
        client,
        timestamp=timestamp,
        is_live=is_live,
    )

    if config.soma_table_query is not None:
        df = df.query(config.soma_table_query)

    if len(df) == 0:
        return None
    elif len(df) == 1:
        return df[config.nucleus_id_column].values[0]
    else:
        return df[config.nucleus_id_column].values



from setuptools import setup
import re
import os
import codecs
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open("requirements.txt", "r") as f:
    required = f.read().splitlines()

setup(
    name="dash-connectivity-viewer",
    version=find_version("dash_connectivity_viewer", "__init__.py"),
    description="Dash connectivity viewer for CAVE data",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Casey Schneider-Mizell",
    author_email="caseysm@gmail.com",
    url="https://github.com/ceesem/dash-connectivity-viewer",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[required],  # external packages as dependencies
)



from dash import dash_table
from dash import dcc
import dash_bootstrap_components as dbc
from dash import html
import flask
from ..common.dash_url_helper import create_component_kwargs, State

title = "Connectivity Viewer"


def make_input_row(state):
    return html.Div(
        [
            dbc.Row(
                [
                    dbc.Col(
                        [
                            html.Div("Cell ID:"),
                            dbc.Input(
                                **create_component_kwargs(
                                    state, id_inner="anno-id", value="", type="text"
                                ),
                            ),
                        ],
                        width={"size": 2, "offset": 1},
                        align="end",
                    ),
                    dbc.Col(
                        [
                            dcc.Dropdown(
                                **create_component_kwargs(
                                    state,
                                    id_inner="id-type",
                                    options=[
                                        {
                                            "label": "Root ID",
                                            "value": "root_id",
                                        },
                                        {
                                            "label": "Nucleus ID",
                                            "value": "nucleus_id",
                                        },
                                    ],
                                    value="root_id",
                                    style={
                                        "margin-left": "12px",
                                        "font-size": "12px",
                                    },
                                    clearable=False,
                                )
                            ),
                        ],
                        width={"size": 1},
                        align="end",
                    ),
                    dbc.Col(
                        [
                            dbc.Checklist(
                                **create_component_kwargs(
                                    state,
                                    id_inner="live-query-toggle",
                                    options=[
                                        {
                                            "label": "Live Query",
                                            "value": 1,
                                        },
                                    ],
                                    value=[
                                        1,
                                    ],
                                    switch=True,
                                    style={
                                        "bottom-margin": "10px",
                                        "font-size": "16px",
                                    },
                                )
                            ),
                        ],
                        width={"size": 1, "offset": 1},
                        align="center",
                    ),
                    dbc.Col(
                        dbc.Button(
                            children="Submit",
                            id="submit-button",
                            color="primary",
                            style={"font-size": "18px"},
                        ),
                        width={"size": 1},
                        align="center",
                    ),
                    dbc.Col(
                        dcc.Loading(
                            id="main-loading",
                            children=html.Div(
                                id="main-loading-placeholder",
                                children="",
                            ),
                            type="default",
                            style={"transform": "scale(0.8)"},
                        ),
                        align="end",
                    ),
                ],
                justify="start",
            ),
            dbc.Row(
                [
                    dbc.Col(
                        [
                            html.Div("Table:"),
                            dcc.Dropdown(
                                **create_component_kwargs(
                                    state,
                                    id_inner="cell-type-table-dropdown",
                                    options=[],
                                    value="",
                                    clearable=True,
                                )
                            ),
                        ],
                        width={"size": 3, "offset": 1},
                        align="end",
                    ),
                ],
                justify="start",
            ),
        ]
    )


def make_message_row(state):
    return dbc.Row(
        dbc.Col(
            dbc.Alert(
                id="message-text",
                children="Please select a root id and press Submit",
                color="info",
            ),
            width=10,
        ),
        justify="center",
    )


def make_table_link_row(state):
    return dbc.Row(
        [
            dbc.Col(
                dbc.Spinner(
                    size="sm",
                    children=html.Div(id="link-loading"),
                ),
                width=1,
                align="center",
            ),
            dbc.Col(
                dbc.Button(
                    children="Table View Neuroglancer Link",
                    id="ngl-link",
                    href="",
                    target="_blank",
                    style={"font-size": "16px"},
                    color="primary",
                    external_link=True,
                    disabled=False,
                ),
                width=3,
                align="start",
            ),
            dbc.Col(
                dbc.Button(
                    id="reset-selection",
                    children="Reset Selection",
                    color="warning",
                    size="sm",
                ),
                width=1,
            ),
            dbc.Col(
                html.A(
                    "Instructions for filtering the table",
                    href="https://dash.plotly.com/datatable/filtering",
                    style={"font-size": "15px"},
                    target="_blank",
                ),
                align="center",
                width={"size": 2, "offset": 3},
            ),
        ],
        justify="start",
    )


def make_data_table_content():
    return html.Div(
        dbc.Row(
            [
                dbc.Col(
                    dash_table.DataTable(
                        id="data-table",
                        columns=[{"name": i, "id": i} for i in [""]],
                        data=[],
                        css=[
                            {
                                "selector": "table",
                                "rule": "table-layout: fixed",
                            }
                        ],
                        style_cell={
                            "height": "auto",
                            "width": "12%",
                            "minWidth": "10%",
                            "maxWidth": "15%",
                            "whiteSpace": "normal",
                            "font-size": "11px",
                        },
                        style_header={
                            "font-size": "12px",
                            "fontWeight": "bold",
                        },
                        sort_action="native",
                        sort_mode="multi",
                        filter_action="native",
                        row_selectable="multi",
                        page_current=0,
                        page_action="native",
                        page_size=30,
                        export_format="csv",
                        export_headers="names",
                    ),
                    width=10,
                ),
            ],
            justify="center",
        )
    )

def make_input_link_tab():
    return [
        dbc.Card(
            [
                dbc.CardBody(
                    [
                        html.H4("All Inputs", className="card-title"),
                        html.Div(
                            children=[
                                dbc.Button(
                                    "Generate Link",
                                    id="all-input-link-button",
                                    color="secondary",
                                    className="d-grid gap-2 col-6 mx-auto",
                                    style={"align-items":"center", "justify-content":"center"}
                                ),
                            ]
                        ),
                        dbc.Spinner(
                            html.Div(
                                "", id="all-input-link", className="card-text"
                            ),
                            size="sm",
                        ),
                    ]
                )
            ]
        ),
        dbc.Col(
            dbc.Card(
                [
                    dbc.CardBody(
                        [
                            html.H4("Grouped Inputs", className="card-title"),
                            html.Div(
                                children=[
                                    dbc.Button(
                                        "Generate Link",
                                        id="cell-typed-input-link-button",
                                        color="secondary",
                                        className="d-grid gap-2 col-6 mx-auto",
                                    ),
                                ]
                            ),
                            dbc.Spinner(
                                html.Div(
                                    "",
                                    id="cell-typed-input-link",
                                    className="card-text",
                                ),
                                size="sm",
                            ),
                        ]
                    )
                ]
            ),
        ),
    ]

def make_output_link_tab():
    return [
        dbc.Card(
            [
                dbc.CardBody(
                    [
                        html.H4("All Outputs", className="card-title"),
                        html.Div(
                            children=[
                                dbc.Button(
                                    "Generate Link",
                                    id="all-output-link-button",
                                    color="secondary",
                                    className="d-grid gap-2 col-6 mx-auto",
                                    style={"align-items":"center", "justify-content":"center"},
                                ),
                            ]
                        ),
                        dbc.Spinner(
                            html.Div(
                                "", id="all-output-link", className="card-text"
                            ),
                            size="sm",
                        ),
                    ]
                )
            ],
            style={"width": "18rem"},
        ),
        dbc.Card(
            [
                dbc.CardBody(
                    [
                        html.H4("Grouped Outputs", className="card-title"),
                        html.Div(
                            children=[
                                dbc.Button(
                                    "Generate Link",
                                    id="cell-typed-output-link-button",
                                    color="secondary",
                                    className="d-grid gap-2 col-6 mx-auto",
                                ),
                            ]
                        ),
                        dbc.Spinner(
                            html.Div(
                                "",
                                id="cell-typed-output-link",
                                className="card-text",
                            ),
                            size="sm",
                        ),
                    ]
                )
            ]
        ),
    ]

def page_layout(state: State = None):
    state = state or {}

    header_text = dbc.Row(
        [
            dbc.Col(
                html.Div(id='header-bar'),
                width={"size": 12},
            ),
        ],
    )
    input_row = make_input_row(state)
    message_row = make_message_row(state)
    top_link = make_table_link_row(state)
    data_table = html.Div(
        [
            dcc.Tabs(
                id="connectivity-tab",
                value="tab-pre",
                children=[
                    dcc.Tab(id="input-tab", label="Input", value="tab-post"),
                    dcc.Tab(id="output-tab", label="Output", value="tab-pre"),
                ],
            ),
            make_data_table_content(),
        ]
    )


    input_tab = make_input_link_tab()
    output_tab = make_output_link_tab()

    cell_links = [
        dbc.Row(
            [
                dbc.Col(
                    dbc.CardGroup(
                        input_tab + output_tab,
                    ),
                    width={"size": 10, "offset": 1},
                ),
            ]
        ),
        dbc.Row(
            dbc.Col(
                dbc.Card(
                    dbc.CardBody(
                        [
                            dbc.Row(
                                [
                                    dbc.Col(
                                        [
                                            html.H5(
                                                "Annotation grouping:",
                                                className="card-title",
                                            ),
                                            dbc.Checklist(
                                                options=[
                                                    {
                                                        "label": "Include No Type",
                                                        "value": 1,
                                                    },
                                                ],
                                                value=[],
                                                switch=True,
                                                style={"font-size": "16px"},
                                                id='no-type-annotation',
                                            ),
                                        ],
                                        width={"size": 2},
                                    ),
                                    dbc.Col(
                                        dcc.Dropdown(
                                            options={},
                                            value="",
                                            id="group-by",
                                            searchable=True,
                                        ),
                                        width={'size': 8},
                                    ),
                                ],
                            ),
                        ]
                    )
                ),
                width={"size": 10, "offset": 1},
            ),
        ),
    ]

    plot_data = [
        dbc.Row(
            [
                dbc.Col(
                    dbc.CardGroup(
                        [
                            dbc.Card(
                                dbc.CardBody(
                                    html.Div(
                                        "",
                                        id="violin-plot",
                                        className="text-center v-100 h-100",
                                    ),
                                )
                            ),
                            dbc.Card(
                                dbc.CardBody(
                                    html.Div("", id="scatter-plot", className="text-center h-100 v-100"),
                                )
                            ),
                            dbc.Card(
                                dbc.CardBody(
                                    html.Div("", id="bar-plot", className="text-center v-100 h-100"),
                                )
                            ),
                        ],
                    className="h-100",
                    ),
                )
            ],
        ),
        dbc.Row(
            dbc.Col(
                dbc.Card(
                    dbc.CardBody(
                        [
                            dbc.Row(
                                [
                                    dbc.Col(
                                        [
                                            html.Div(
                                                "Color by value:",
                                                style={'align-content': 'right'},
                                            ),
                                        ],
                                        style={'align-content': 'right'},
                                        width={"size": 1, 'offset':1},
                                    ),
                                    dbc.Col(
                                        dcc.Dropdown(
                                            options={"cell_type": "cell_type"},
                                            value="cell_type",
                                            id="plot-color-value",
                                        ),  
                                        style={"align-content": "left"},
                                        width={'size': 9},
                                    ),
                                ],
                            ),
                        ]
                    )
                ),
                # width={"size": 12},
            ),
        ),
    ]

    main_tab = dbc.Row(
        [
            dbc.Col(
                dbc.Tabs(
                    [
                        dbc.Tab(html.Div(), label="Table Only"),
                        dbc.Tab(plot_data, label="Plots"),
                        dbc.Tab(cell_links, label="Neuroglancer Links")
                    ],
                ),
                width={'size': 10, 'offset': 1},
            )
        ]
    )

    layout = html.Div(
        children=[
            header_text,
            dbc.Container(input_row, fluid=True),
            html.Br(),
            html.Div(message_row),
            html.Hr(),
            html.Div(main_tab),
            # html.Div(plot_data),
            # html.Hr(),
            # dbc.Container(cell_links),
            html.Hr(),
            top_link,
            data_table,
            dcc.Store("target-table-json"),
            dcc.Store("source-table-json"),
            dcc.Store("client-info-json"),
            dcc.Store("synapse-table-resolution-json"),
            dcc.Store("value-columns"),
            html.Div(
                dcc.Input(
                    **create_component_kwargs(
                        state,
                        id_inner="datastack",
                        value="",
                    ),
                ),
                style={"display": "none"},
            ),
        ]
    )

    return layout


#################
# URL-formatter #
#################

url_bar_and_content_div = html.Div(
    [dcc.Location(id="url", refresh=False), html.Div(id="page-layout")]
)


def app_layout():
    # https://dash.plotly.com/urls "Dynamically Create a Layout for Multi-Page App Validation"
    if flask.has_request_context():  # for real
        return url_bar_and_content_div
    # validation only
    return html.Div([url_bar_and_content_div, *page_layout()])
